#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2019, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
# PySNMP MIB module SNMP-FRAMEWORK-MIB (https://www.pysnmp.com/pysnmp)
# ASN.1 source http://mibs.pysnmp.com:80/asn1/SNMP-FRAMEWORK-MIB
# Produced by pysmi-0.1.3 at Mon Apr 17 13:59:41 2017
# On host grommit.local platform Darwin version 16.4.0 by user ilya
# Using Python version 3.4.2 (v3.4.2:ab2c023a9432, Oct  5 2014, 20:42:22)
#
import platform
import time
try:
    import os
except ImportError:
    pass

OctetString, ObjectIdentifier, Integer = mibBuilder.importSymbols("ASN1", "OctetString", "ObjectIdentifier", "Integer")
NamedValues, = mibBuilder.importSymbols("ASN1-ENUMERATION", "NamedValues")
ValueSizeConstraint, SingleValueConstraint, ValueRangeConstraint, ConstraintsUnion, ConstraintsIntersection = mibBuilder.importSymbols("ASN1-REFINEMENT", "ValueSizeConstraint", "SingleValueConstraint", "ValueRangeConstraint", "ConstraintsUnion", "ConstraintsIntersection")
ObjectGroup, NotificationGroup, ModuleCompliance = mibBuilder.importSymbols("SNMPv2-CONF", "ObjectGroup", "NotificationGroup", "ModuleCompliance")
Gauge32, Unsigned32, MibIdentifier, ObjectIdentity, iso, Counter64, IpAddress, Counter32, ModuleIdentity, Bits, snmpModules, Integer32, MibScalar, MibTable, MibTableRow, MibTableColumn, TimeTicks, NotificationType = mibBuilder.importSymbols("SNMPv2-SMI", "Gauge32", "Unsigned32", "MibIdentifier", "ObjectIdentity", "iso", "Counter64", "IpAddress", "Counter32", "ModuleIdentity", "Bits", "snmpModules", "Integer32", "MibScalar", "MibTable", "MibTableRow", "MibTableColumn", "TimeTicks", "NotificationType")
DisplayString, TextualConvention = mibBuilder.importSymbols("SNMPv2-TC", "DisplayString", "TextualConvention")
snmpFrameworkMIB = ModuleIdentity((1, 3, 6, 1, 6, 3, 10))
if mibBuilder.loadTexts: snmpFrameworkMIB.setRevisions(('2002-10-14 00:00', '1999-01-19 00:00', '1997-11-20 00:00',))
if mibBuilder.loadTexts: snmpFrameworkMIB.setLastUpdated('200210140000Z')
if mibBuilder.loadTexts: snmpFrameworkMIB.setOrganization('SNMPv3 Working Group')
if mibBuilder.loadTexts: snmpFrameworkMIB.setContactInfo('WG-EMail: snmpv3@lists.tislabs.com Subscribe: snmpv3-request@lists.tislabs.com Co-Chair: Russ Mundy Network Associates Laboratories postal: 15204 Omega Drive, Suite 300 Rockville, MD 20850-4601 USA EMail: mundy@tislabs.com phone: +1 301-947-7107 Co-Chair & Co-editor: David Harrington Enterasys Networks postal: 35 Industrial Way P. O. Box 5005 Rochester, New Hampshire 03866-5005 USA EMail: dbh@enterasys.com phone: +1 603-337-2614 Co-editor: Randy Presuhn BMC Software, Inc. postal: 2141 North First Street San Jose, California 95131 USA EMail: randy_presuhn@bmc.com phone: +1 408-546-1006 Co-editor: Bert Wijnen Lucent Technologies postal: Schagen 33 3461 GL Linschoten Netherlands EMail: bwijnen@lucent.com phone: +31 348-680-485 ')
if mibBuilder.loadTexts: snmpFrameworkMIB.setDescription('The SNMP Management Architecture MIB Copyright (C) The Internet Society (2002). This version of this MIB module is part of RFC 3411; see the RFC itself for full legal notices. ')

#
# WARNING: some of the classes below are manually implemented
#

class SnmpAdminString(TextualConvention, OctetString):
    description = 'An octet string containing administrative information, preferably in human-readable form. To facilitate internationalization, this information is represented using the ISO/IEC IS 10646-1 character set, encoded as an octet string using the UTF-8 transformation format described in [RFC2279]. Since additional code points are added by amendments to the 10646 standard from time to time, implementations must be prepared to encounter any code point from 0x00000000 to 0x7fffffff. Byte sequences that do not correspond to the valid UTF-8 encoding of a code point or are outside this range are prohibited. The use of control codes should be avoided. When it is necessary to represent a newline, the control code sequence CR LF should be used. The use of leading or trailing white space should be avoided. For code points not directly supported by user interface hardware or software, an alternative means of entry and display, such as hexadecimal, may be provided. For information encoded in 7-bit US-ASCII, the UTF-8 encoding is identical to the US-ASCII encoding. UTF-8 may require multiple bytes to represent a single character / code point; thus the length of this object in octets may be different from the number of characters encoded. Similarly, size constraints refer to the number of encoded octets, not the number of characters represented by an encoding. Note that when this TC is used for an object that is used or envisioned to be used as an index, then a SIZE restriction MUST be specified so that the number of sub-identifiers for any object instance does not exceed the limit of 128, as defined by [RFC3416]. Note that the size of an SnmpAdminString object is measured in octets, not characters. '
    status = 'current'
    displayHint = "255t"
    subtypeSpec = OctetString.subtypeSpec + ValueSizeConstraint(0, 255)
    encoding = 'utf-8'


class SnmpEngineID(TextualConvention, OctetString):
    description = "An SNMP engine's administratively-unique identifier. Objects of this type are for identification, not for addressing, even though it is possible that an address may have been used in the generation of a specific value. The value for this object may not be all zeros or all 'ff'H or the empty (zero length) string. The initial value for this object may be configured via an operator console entry or via an algorithmic function. In the latter case, the following example algorithm is recommended. In cases where there are multiple engines on the same system, the use of this algorithm is NOT appropriate, as it would result in all of those engines ending up with the same ID value. 1) The very first bit is used to indicate how the rest of the data is composed. 0 - as defined by enterprise using former methods that existed before SNMPv3. See item 2 below. 1 - as defined by this architecture, see item 3 below. Note that this allows existing uses of the engineID (also known as AgentID [RFC1910]) to co-exist with any new uses. 2) The snmpEngineID has a length of 12 octets. The first four octets are set to the binary equivalent of the agent's SNMP management private enterprise number as assigned by the Internet Assigned Numbers Authority (IANA). For example, if Acme Networks has been assigned { enterprises 696 }, the first four octets would be assigned '000002b8'H. The remaining eight octets are determined via one or more enterprise-specific methods. Such methods must be designed so as to maximize the possibility that the value of this object will be unique in the agent's administrative domain. For example, it may be the IP address of the SNMP entity, or the MAC address of one of the interfaces, with each address suitably padded with random octets. If multiple methods are defined, then it is recommended that the first octet indicate the method being used and the remaining octets be a function of the method. 3) The length of the octet string varies. The first four octets are set to the binary equivalent of the agent's SNMP management private enterprise number as assigned by the Internet Assigned Numbers Authority (IANA). For example, if Acme Networks has been assigned { enterprises 696 }, the first four octets would be assigned '000002b8'H. The very first bit is set to 1. For example, the above value for Acme Networks now changes to be '800002b8'H. The fifth octet indicates how the rest (6th and following octets) are formatted. The values for the fifth octet are: 0 - reserved, unused. 1 - IPv4 address (4 octets) lowest non-special IP address 2 - IPv6 address (16 octets) lowest non-special IP address 3 - MAC address (6 octets) lowest IEEE MAC address, canonical order 4 - Text, administratively assigned Maximum remaining length 27 5 - Octets, administratively assigned Maximum remaining length 27 6-127 - reserved, unused 128-255 - as defined by the enterprise Maximum remaining length 27 "
    status = 'current'
    subtypeSpec = OctetString.subtypeSpec + ValueSizeConstraint(5, 32)
    defaultValue = [128, 0, 79, 184, 5]
    try:
        # Attempt to base engine ID on local system name and properties
        defaultValue += [ord(x) for x in platform.node()[:16]]
    except Exception:
        pass
    try:
        # Attempt to base engine ID on PID
        defaultValue += [os.getpid() >> 8 & 0xff, os.getpid() & 0xff]
    except Exception:
        pass
    # add pseudo-random text ID
    defaultValue += [id(defaultValue) >> 8 & 0xff, id(defaultValue) & 0xff]
    defaultValue = OctetString(defaultValue).asOctets()


class SnmpEngineTime(Integer32):
    def clone(self, *args, **kwargs):
        if not args:
            try:
                args = (time.time() - self,)
            except Exception:
                pass
        return Integer32.clone(self, *args, **kwargs)


class SnmpSecurityModel(Integer32, TextualConvention):
    description = "An identifier that uniquely identifies a Security Model of the Security Subsystem within this SNMP Management Architecture. The values for securityModel are allocated as follows: - The zero value does not identify any particular security model. - Values between 1 and 255, inclusive, are reserved for standards-track Security Models and are managed by the Internet Assigned Numbers Authority (IANA). - Values greater than 255 are allocated to enterprise-specific Security Models. An enterprise-specific securityModel value is defined to be: enterpriseID * 256 + security model within enterprise For example, the fourth Security Model defined by the enterprise whose enterpriseID is 1 would be 259. This scheme for allocation of securityModel values allows for a maximum of 255 standards- based Security Models, and for a maximum of 256 Security Models per enterprise. It is believed that the assignment of new securityModel values will be rare in practice because the larger the number of simultaneously utilized Security Models, the larger the chance that interoperability will suffer. Consequently, it is believed that such a range will be sufficient. In the unlikely event that the standards committee finds this number to be insufficient over time, an enterprise number can be allocated to obtain an additional 256 possible values. Note that the most significant bit must be zero; hence, there are 23 bits allocated for various organizations to design and define non-standard securityModels. This limits the ability to define new proprietary implementations of Security Models to the first 8,388,608 enterprises. It is worthwhile to note that, in its encoded form, the securityModel value will normally require only a single byte since, in practice, the leftmost bits will be zero for most messages and sign extension is suppressed by the encoding rules. As of this writing, there are several values of securityModel defined for use with SNMP or reserved for use with supporting MIB objects. They are as follows: 0 reserved for 'any' 1 reserved for SNMPv1 2 reserved for SNMPv2c 3 User-Based Security Model (USM) "
    status = 'current'
    subtypeSpec = Integer32.subtypeSpec + ValueRangeConstraint(0, 2147483647)


class SnmpMessageProcessingModel(Integer32, TextualConvention):
    description = 'An identifier that uniquely identifies a Message Processing Model of the Message Processing Subsystem within this SNMP Management Architecture. The values for messageProcessingModel are allocated as follows: - Values between 0 and 255, inclusive, are reserved for standards-track Message Processing Models and are managed by the Internet Assigned Numbers Authority (IANA). - Values greater than 255 are allocated to enterprise-specific Message Processing Models. An enterprise messageProcessingModel value is defined to be: enterpriseID * 256 + messageProcessingModel within enterprise For example, the fourth Message Processing Model defined by the enterprise whose enterpriseID is 1 would be 259. This scheme for allocating messageProcessingModel values allows for a maximum of 255 standards- based Message Processing Models, and for a maximum of 256 Message Processing Models per enterprise. It is believed that the assignment of new messageProcessingModel values will be rare in practice because the larger the number of simultaneously utilized Message Processing Models, the larger the chance that interoperability will suffer. It is believed that such a range will be sufficient. In the unlikely event that the standards committee finds this number to be insufficient over time, an enterprise number can be allocated to obtain an additional 256 possible values. Note that the most significant bit must be zero; hence, there are 23 bits allocated for various organizations to design and define non-standard messageProcessingModels. This limits the ability to define new proprietary implementations of Message Processing Models to the first 8,388,608 enterprises. It is worthwhile to note that, in its encoded form, the messageProcessingModel value will normally require only a single byte since, in practice, the leftmost bits will be zero for most messages and sign extension is suppressed by the encoding rules. As of this writing, there are several values of messageProcessingModel defined for use with SNMP. They are as follows: 0 reserved for SNMPv1 1 reserved for SNMPv2c 2 reserved for SNMPv2u and SNMPv2* 3 reserved for SNMPv3 '
    status = 'current'
    subtypeSpec = Integer32.subtypeSpec + ValueRangeConstraint(0, 2147483647)


class SnmpSecurityLevel(Integer32, TextualConvention):
    description = 'A Level of Security at which SNMP messages can be sent or with which operations are being processed; in particular, one of: noAuthNoPriv - without authentication and without privacy, authNoPriv - with authentication but without privacy, authPriv - with authentication and with privacy. These three values are ordered such that noAuthNoPriv is less than authNoPriv and authNoPriv is less than authPriv. '
    status = 'current'
    subtypeSpec = Integer32.subtypeSpec + SingleValueConstraint(1, 2, 3, )
    namedValues = NamedValues(("noAuthNoPriv", 1), ("authNoPriv", 2), ("authPriv", 3), )

snmpFrameworkAdmin = MibIdentifier((1, 3, 6, 1, 6, 3, 10, 1))
snmpFrameworkMIBObjects = MibIdentifier((1, 3, 6, 1, 6, 3, 10, 2))
snmpFrameworkMIBConformance = MibIdentifier((1, 3, 6, 1, 6, 3, 10, 3))
snmpEngine = MibIdentifier((1, 3, 6, 1, 6, 3, 10, 2, 1))
snmpEngineID = MibScalar((1, 3, 6, 1, 6, 3, 10, 2, 1, 1), SnmpEngineID()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpEngineID.setStatus('current')
if mibBuilder.loadTexts: snmpEngineID.setDescription("An SNMP engine's administratively-unique identifier. This information SHOULD be stored in non-volatile storage so that it remains constant across re-initializations of the SNMP engine. ")
snmpEngineBoots = MibScalar((1, 3, 6, 1, 6, 3, 10, 2, 1, 2), Integer32().subtype(subtypeSpec=ValueRangeConstraint(1, 2147483647))).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpEngineBoots.setStatus('current')
if mibBuilder.loadTexts: snmpEngineBoots.setDescription('The number of times that the SNMP engine has (re-)initialized itself since snmpEngineID was last configured. ')
snmpEngineTime = MibScalar((1, 3, 6, 1, 6, 3, 10, 2, 1, 3), SnmpEngineTime().subtype(subtypeSpec=ValueRangeConstraint(0, 2147483647))).setUnits('seconds').setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpEngineTime.setStatus('current')
if mibBuilder.loadTexts: snmpEngineTime.setDescription("The number of seconds since the value of the snmpEngineBoots object last changed. When incrementing this object's value would cause it to exceed its maximum, snmpEngineBoots is incremented as if a re-initialization had occurred, and this object's value consequently reverts to zero. ")
snmpEngineMaxMessageSize = MibScalar((1, 3, 6, 1, 6, 3, 10, 2, 1, 4), Integer32().subtype(subtypeSpec=ValueRangeConstraint(484, 2147483647))).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpEngineMaxMessageSize.setStatus('current')
if mibBuilder.loadTexts: snmpEngineMaxMessageSize.setDescription('The maximum length in octets of an SNMP message which this SNMP engine can send or receive and process, determined as the minimum of the maximum message size values supported among all of the transports available to and supported by the engine. ')
snmpAuthProtocols = ObjectIdentity((1, 3, 6, 1, 6, 3, 10, 1, 1))
if mibBuilder.loadTexts: snmpAuthProtocols.setStatus('current')
if mibBuilder.loadTexts: snmpAuthProtocols.setDescription('Registration point for standards-track authentication protocols used in SNMP Management Frameworks. ')
snmpPrivProtocols = ObjectIdentity((1, 3, 6, 1, 6, 3, 10, 1, 2))
if mibBuilder.loadTexts: snmpPrivProtocols.setStatus('current')
if mibBuilder.loadTexts: snmpPrivProtocols.setDescription('Registration point for standards-track privacy protocols used in SNMP Management Frameworks. ')
snmpFrameworkMIBCompliances = MibIdentifier((1, 3, 6, 1, 6, 3, 10, 3, 1))
snmpFrameworkMIBGroups = MibIdentifier((1, 3, 6, 1, 6, 3, 10, 3, 2))
snmpFrameworkMIBCompliance = ModuleCompliance((1, 3, 6, 1, 6, 3, 10, 3, 1, 1)).setObjects(("SNMP-FRAMEWORK-MIB", "snmpEngineGroup"))
if mibBuilder.loadTexts: snmpFrameworkMIBCompliance.setDescription('The compliance statement for SNMP engines which implement the SNMP Management Framework MIB. ')
snmpEngineGroup = ObjectGroup((1, 3, 6, 1, 6, 3, 10, 3, 2, 1)).setObjects(("SNMP-FRAMEWORK-MIB", "snmpEngineID"), ("SNMP-FRAMEWORK-MIB", "snmpEngineBoots"), ("SNMP-FRAMEWORK-MIB", "snmpEngineTime"), ("SNMP-FRAMEWORK-MIB", "snmpEngineMaxMessageSize"))
if mibBuilder.loadTexts: snmpEngineGroup.setDescription('A collection of objects for identifying and determining the configuration and current timeliness values of an SNMP engine. ')
mibBuilder.exportSymbols("SNMP-FRAMEWORK-MIB", snmpFrameworkMIBConformance=snmpFrameworkMIBConformance, snmpFrameworkMIB=snmpFrameworkMIB, snmpFrameworkMIBGroups=snmpFrameworkMIBGroups, SnmpEngineID=SnmpEngineID, SnmpSecurityLevel=SnmpSecurityLevel, snmpEngineMaxMessageSize=snmpEngineMaxMessageSize, snmpFrameworkAdmin=snmpFrameworkAdmin, PYSNMP_MODULE_ID=snmpFrameworkMIB, snmpFrameworkMIBObjects=snmpFrameworkMIBObjects, snmpFrameworkMIBCompliances=snmpFrameworkMIBCompliances, snmpAuthProtocols=snmpAuthProtocols, snmpFrameworkMIBCompliance=snmpFrameworkMIBCompliance, snmpEngineTime=snmpEngineTime, snmpPrivProtocols=snmpPrivProtocols, SnmpMessageProcessingModel=SnmpMessageProcessingModel, snmpEngine=snmpEngine, SnmpSecurityModel=SnmpSecurityModel, snmpEngineGroup=snmpEngineGroup, snmpEngineBoots=snmpEngineBoots, SnmpAdminString=SnmpAdminString, snmpEngineID=snmpEngineID)



from pysnmp.entity.engine import SnmpEngine
from pysnmp.hlapi.asyncio.sync.cmdgen import walkCmd
from pysnmp.hlapi.asyncio.transport import UdpTransportTarget
from pysnmp.hlapi.auth import CommunityData
from pysnmp.hlapi.context import ContextData
from pysnmp.proto.rfc1155 import TimeTicks
from pysnmp.proto.rfc1902 import OctetString
from pysnmp.smi import compiler
from pysnmp.smi.rfc1902 import ObjectIdentity, ObjectType


def test_v1_walk():
    snmpEngine = SnmpEngine()

    builder = snmpEngine.getMibBuilder()
    # Attach MIB compiler to SNMP Engine (MIB Builder)
    # This call will fail if PySMI is not present on the system
    compiler.addMibCompiler(builder)
    # ... alternatively, this call will not complain on missing PySMI
    # compiler.addMibCompiler(snmpEngine.getMibBuilder(), ifAvailable=True)

    builder.loadModules("IF-MIB")

    objects = walkCmd(
        snmpEngine,
        CommunityData("public", mpModel=0),
        UdpTransportTarget(("demo.pysnmp.com", 161)),
        ContextData(),
        ObjectType(ObjectIdentity("SNMPv2-MIB", "sysDescr", 0)),
    )

    errorIndication, errorStatus, errorIndex, varBinds = next(objects)

    assert errorIndication is None
    assert errorStatus == 0
    assert len(varBinds) == 1
    assert varBinds[0][0].prettyPrint() == "SNMPv2-MIB::sysObjectID.0"
    assert varBinds[0][1].prettyPrint() == "SNMPv2-SMI::internet"
    # assert isinstance(varBinds[0][1], ObjectIdentifier)

    errorIndication, errorStatus, errorIndex, varBinds = next(objects)

    assert errorIndication is None
    assert errorStatus == 0
    assert len(varBinds) == 1
    assert varBinds[0][0].prettyPrint() == "SNMPv2-MIB::sysUpTime.0"
    # assert isinstance(varBinds[0][1], TimeTicks)

    objects_list = list(objects)
    assert len(objects_list), 50

    errorIndication, errorStatus, errorIndex, varBinds = objects_list[-1]
    assert varBinds[0][0].prettyPrint() == "IF-MIB::ifSpecific.2"

    snmpEngine.transportDispatcher.closeDispatcher()



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
#
# Copyright (c) 2022-2024, LeXtudio Inc. <support@lextudio.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
# All code in this file belongs to obsolete, compatibility wrappers.
# Never use interfaces below for new applications!
#
from pysnmp.hlapi.asyncio import *
from pysnmp.hlapi.asyncio import sync
from pysnmp.hlapi.varbinds import *
from pysnmp.hlapi.lcd import *
from pyasn1.compat.octets import null
from pyasn1.type import univ

__all__ = ["CommandGenerator", "MibVariable"]

MibVariable = ObjectIdentity


class CommandGenerator:
    _null = univ.Null("")

    def __init__(self, snmpEngine=None, asynCmdGen=None):
        # compatibility attributes
        self.snmpEngine = snmpEngine or SnmpEngine()

    def getCmd(self, authData, transportTarget, *varNames, **kwargs):
        if "lookupNames" not in kwargs:
            kwargs["lookupNames"] = False
        if "lookupValues" not in kwargs:
            kwargs["lookupValues"] = False
        return sync.getCmd(
            self.snmpEngine,
            authData,
            transportTarget,
            ContextData(kwargs.get("contextEngineId"), kwargs.get("contextName", null)),
            *[(x, self._null) for x in varNames],
            **kwargs
        )

    def setCmd(self, authData, transportTarget, *varBinds, **kwargs):
        if "lookupNames" not in kwargs:
            kwargs["lookupNames"] = False
        if "lookupValues" not in kwargs:
            kwargs["lookupValues"] = False
        return sync.setCmd(
            self.snmpEngine,
            authData,
            transportTarget,
            ContextData(kwargs.get("contextEngineId"), kwargs.get("contextName", null)),
            *varBinds,
            **kwargs
        )

    def nextCmd(self, authData, transportTarget, *varNames, **kwargs):
        if "lookupNames" not in kwargs:
            kwargs["lookupNames"] = False
        if "lookupValues" not in kwargs:
            kwargs["lookupValues"] = False
        if "lexicographicMode" not in kwargs:
            kwargs["lexicographicMode"] = False
        errorIndication, errorStatus, errorIndex = None, 0, 0
        varBindTable = []
        for errorIndication, errorStatus, errorIndex, varBinds in sync.walkCmd(
            self.snmpEngine,
            authData,
            transportTarget,
            ContextData(kwargs.get("contextEngineId"), kwargs.get("contextName", null)),
            *[(x, self._null) for x in varNames],
            **kwargs
        ):
            if errorIndication or errorStatus:
                return errorIndication, errorStatus, errorIndex, varBinds

            varBindTable.append(varBinds)

        return errorIndication, errorStatus, errorIndex, varBindTable

    def bulkCmd(
        self,
        authData,
        transportTarget,
        nonRepeaters,
        maxRepetitions,
        *varNames,
        **kwargs
    ):
        if "lookupNames" not in kwargs:
            kwargs["lookupNames"] = False
        if "lookupValues" not in kwargs:
            kwargs["lookupValues"] = False
        if "lexicographicMode" not in kwargs:
            kwargs["lexicographicMode"] = False
        errorIndication, errorStatus, errorIndex = None, 0, 0
        varBindTable = []
        for errorIndication, errorStatus, errorIndex, varBinds in sync.bulkWalkCmd(
            self.snmpEngine,
            authData,
            transportTarget,
            ContextData(kwargs.get("contextEngineId"), kwargs.get("contextName", null)),
            nonRepeaters,
            maxRepetitions,
            *[(x, self._null) for x in varNames],
            **kwargs
        ):
            if errorIndication or errorStatus:
                return errorIndication, errorStatus, errorIndex, varBinds

            varBindTable.append(varBinds)

        return errorIndication, errorStatus, errorIndex, varBindTable



from pysnmp.proto.rfc1902 import *
from pysnmp.smi.rfc1902 import *
from pysnmp.hlapi.auth import *
from pysnmp.hlapi.context import *
from pysnmp.hlapi.asyncio.transport import *
from pysnmp.entity.engine import *


from pysnmp.hlapi.asyncio.sync.cmdgen import *
from pysnmp.hlapi.asyncio.sync.ntforg import *



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2019, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
#
# PySNMP MIB module SNMP-MPD-MIB (https://www.pysnmp.com/pysnmp)
# ASN.1 source http://mibs.pysnmp.com:80/asn1/SNMP-MPD-MIB
# Produced by pysmi-0.1.3 at Tue Apr 18 00:17:00 2017
# On host grommit.local platform Darwin version 16.4.0 by user ilya
# Using Python version 3.4.2 (v3.4.2:ab2c023a9432, Oct  5 2014, 20:42:22)
#
OctetString, ObjectIdentifier, Integer = mibBuilder.importSymbols("ASN1", "OctetString", "ObjectIdentifier", "Integer")
NamedValues, = mibBuilder.importSymbols("ASN1-ENUMERATION", "NamedValues")
SingleValueConstraint, ConstraintsUnion, ValueSizeConstraint, ValueRangeConstraint, ConstraintsIntersection = mibBuilder.importSymbols("ASN1-REFINEMENT", "SingleValueConstraint", "ConstraintsUnion", "ValueSizeConstraint", "ValueRangeConstraint", "ConstraintsIntersection")
NotificationGroup, ModuleCompliance, ObjectGroup = mibBuilder.importSymbols("SNMPv2-CONF", "NotificationGroup", "ModuleCompliance", "ObjectGroup")
iso, Counter32, Unsigned32, Gauge32, MibIdentifier, TimeTicks, NotificationType, ObjectIdentity, Bits, ModuleIdentity, MibScalar, MibTable, MibTableRow, MibTableColumn, Integer32, Counter64, IpAddress, snmpModules = mibBuilder.importSymbols("SNMPv2-SMI", "iso", "Counter32", "Unsigned32", "Gauge32", "MibIdentifier", "TimeTicks", "NotificationType", "ObjectIdentity", "Bits", "ModuleIdentity", "MibScalar", "MibTable", "MibTableRow", "MibTableColumn", "Integer32", "Counter64", "IpAddress", "snmpModules")
DisplayString, TextualConvention = mibBuilder.importSymbols("SNMPv2-TC", "DisplayString", "TextualConvention")
snmpMPDMIB = ModuleIdentity((1, 3, 6, 1, 6, 3, 11))
if mibBuilder.loadTexts: snmpMPDMIB.setRevisions(('1999-05-04 16:36', '1997-09-30 00:00',))
if mibBuilder.loadTexts: snmpMPDMIB.setLastUpdated('9905041636Z')
if mibBuilder.loadTexts: snmpMPDMIB.setOrganization('SNMPv3 Working Group')
if mibBuilder.loadTexts: snmpMPDMIB.setContactInfo('WG-EMail: snmpv3@lists.tislabs.com Subscribe: majordomo@lists.tislabs.com In message body: subscribe snmpv3 Chair: Russ Mundy TIS Labs at Network Associates postal: 3060 Washington Road Glenwood, MD 21738 USA EMail: mundy@tislabs.com phone: +1 301-854-6889 Co-editor: Jeffrey Case SNMP Research, Inc. postal: 3001 Kimberlin Heights Road Knoxville, TN 37920-9716 USA EMail: case@snmp.com phone: +1 423-573-1434 Co-editor Dave Harrington Cabletron Systems, Inc. postal: Post Office Box 5005 MailStop: Durham 35 Industrial Way Rochester, NH 03867-5005 USA EMail: dbh@ctron.com phone: +1 603-337-7357 Co-editor: Randy Presuhn BMC Software, Inc. postal: 965 Stewart Drive Sunnyvale, CA 94086 USA EMail: randy_presuhn@bmc.com phone: +1 408-616-3100 Co-editor: Bert Wijnen IBM T. J. Watson Research postal: Schagen 33 3461 GL Linschoten Netherlands EMail: wijnen@vnet.ibm.com phone: +31 348-432-794 ')
if mibBuilder.loadTexts: snmpMPDMIB.setDescription('The MIB for Message Processing and Dispatching')
snmpMPDAdmin = MibIdentifier((1, 3, 6, 1, 6, 3, 11, 1))
snmpMPDMIBObjects = MibIdentifier((1, 3, 6, 1, 6, 3, 11, 2))
snmpMPDMIBConformance = MibIdentifier((1, 3, 6, 1, 6, 3, 11, 3))
snmpMPDStats = MibIdentifier((1, 3, 6, 1, 6, 3, 11, 2, 1))
snmpUnknownSecurityModels = MibScalar((1, 3, 6, 1, 6, 3, 11, 2, 1, 1), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpUnknownSecurityModels.setStatus('current')
if mibBuilder.loadTexts: snmpUnknownSecurityModels.setDescription('The total number of packets received by the SNMP engine which were dropped because they referenced a securityModel that was not known to or supported by the SNMP engine. ')
snmpInvalidMsgs = MibScalar((1, 3, 6, 1, 6, 3, 11, 2, 1, 2), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpInvalidMsgs.setStatus('current')
if mibBuilder.loadTexts: snmpInvalidMsgs.setDescription('The total number of packets received by the SNMP engine which were dropped because there were invalid or inconsistent components in the SNMP message. ')
snmpUnknownPDUHandlers = MibScalar((1, 3, 6, 1, 6, 3, 11, 2, 1, 3), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpUnknownPDUHandlers.setStatus('current')
if mibBuilder.loadTexts: snmpUnknownPDUHandlers.setDescription('The total number of packets received by the SNMP engine which were dropped because the PDU contained in the packet could not be passed to an application responsible for handling the pduType, e.g. no SNMP application had registered for the proper combination of the contextEngineID and the pduType. ')
snmpMPDMIBCompliances = MibIdentifier((1, 3, 6, 1, 6, 3, 11, 3, 1))
snmpMPDMIBGroups = MibIdentifier((1, 3, 6, 1, 6, 3, 11, 3, 2))
snmpMPDCompliance = ModuleCompliance((1, 3, 6, 1, 6, 3, 11, 3, 1, 1)).setObjects(("SNMP-MPD-MIB", "snmpMPDGroup"))
if mibBuilder.loadTexts: snmpMPDCompliance.setDescription('The compliance statement for SNMP entities which implement the SNMP-MPD-MIB. ')
snmpMPDGroup = ObjectGroup((1, 3, 6, 1, 6, 3, 11, 3, 2, 1)).setObjects(("SNMP-MPD-MIB", "snmpUnknownSecurityModels"), ("SNMP-MPD-MIB", "snmpInvalidMsgs"), ("SNMP-MPD-MIB", "snmpUnknownPDUHandlers"))
if mibBuilder.loadTexts: snmpMPDGroup.setDescription('A collection of objects providing for remote monitoring of the SNMP Message Processing and Dispatching process. ')
mibBuilder.exportSymbols("SNMP-MPD-MIB", snmpMPDMIBGroups=snmpMPDMIBGroups, PYSNMP_MODULE_ID=snmpMPDMIB, snmpMPDMIBConformance=snmpMPDMIBConformance, snmpMPDMIBCompliances=snmpMPDMIBCompliances, snmpUnknownPDUHandlers=snmpUnknownPDUHandlers, snmpMPDGroup=snmpMPDGroup, snmpMPDMIB=snmpMPDMIB, snmpInvalidMsgs=snmpInvalidMsgs, snmpMPDCompliance=snmpMPDCompliance, snmpMPDMIBObjects=snmpMPDMIBObjects, snmpMPDAdmin=snmpMPDAdmin, snmpMPDStats=snmpMPDStats, snmpUnknownSecurityModels=snmpUnknownSecurityModels)



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2019, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
from pysnmp.smi.error import NoSuchInstanceError
from pysnmp.proto import errind, error
from pysnmp import debug


# 3.2
class Vacm:
    """View-based Access Control Model"""

    accessModelID = 3

    _powOfTwoSeq = (128, 64, 32, 16, 8, 4, 2, 1)

    def __init__(self):
        self._contextBranchId = -1
        self._groupNameBranchId = -1
        self._accessBranchId = -1
        self._viewTreeBranchId = -1

        self._contextMap = {}
        self._groupNameMap = {}
        self._accessMap = {}
        self._viewTreeMap = {}

    def _addAccessEntry(
        self,
        groupName,
        contextPrefix,
        securityModel,
        securityLevel,
        prefixMatch,
        readView,
        writeView,
        notifyView,
    ):
        if not groupName:
            return

        groups = self._accessMap

        try:
            views = groups[groupName]

        except KeyError:
            views = groups[groupName] = {}

        for viewType, viewName in (
            ("read", readView),
            ("write", writeView),
            ("notify", notifyView),
        ):
            try:
                matches = views[viewType]

            except KeyError:
                matches = views[viewType] = {}

            try:
                contexts = matches[prefixMatch]

            except KeyError:
                contexts = matches[prefixMatch] = {}

            try:
                models = contexts[contextPrefix]

            except KeyError:
                models = contexts[contextPrefix] = {}

            try:
                levels = models[securityModel]

            except KeyError:
                levels = models[securityModel] = {}

            levels[securityLevel] = viewName

    def _getFamilyViewName(
        self, groupName, contextName, securityModel, securityLevel, viewType
    ):
        groups = self._accessMap

        try:
            views = groups[groupName]

        except KeyError:
            raise error.StatusInformation(errorIndication=errind.noGroupName)

        try:
            matches = views[viewType]

        except KeyError:
            raise error.StatusInformation(errorIndication=errind.noAccessEntry)

        try:
            # vacmAccessTable #2: exact match shortcut
            return matches[1][contextName][securityModel][securityLevel]

        except KeyError:
            pass

        # vacmAccessTable #2: fuzzy look-up

        candidates = []

        for match, names in matches.items():
            for context, models in names.items():
                if match == 1 and contextName != context:
                    continue

                if match == 2 and contextName[: len(context)] != context:
                    continue

                for model, levels in models.items():
                    for level, viewName in levels.items():
                        # priorities:
                        # - matching securityModel
                        # - exact context name match
                        # - longer partial match
                        # - highest securityLevel
                        rating = securityModel == model, match == 1, len(context), level

                        candidates.append((rating, viewName))

        if not candidates:
            raise error.StatusInformation(errorIndication=errind.notInView)

        candidates.sort()

        rating, viewName = candidates[0]
        return viewName

    def isAccessAllowed(
        self,
        snmpEngine,
        securityModel,
        securityName,
        securityLevel,
        viewType,
        contextName,
        variableName,
    ):
        mibInstrumController = snmpEngine.msgAndPduDsp.mibInstrumController

        debug.logger & debug.flagACL and debug.logger(
            "isAccessAllowed: securityModel %s, securityName %s, "
            "securityLevel %s, viewType %s, contextName %s for "
            "variableName %s"
            % (
                securityModel,
                securityName,
                securityLevel,
                viewType,
                contextName,
                variableName,
            )
        )

        # Rebuild contextName map if changed

        (vacmContextName,) = mibInstrumController.mibBuilder.importSymbols(
            "SNMP-VIEW-BASED-ACM-MIB", "vacmContextName"
        )

        if self._contextBranchId != vacmContextName.branchVersionId:
            self._contextMap.clear()

            nextMibNode = vacmContextName

            while True:
                try:
                    nextMibNode = vacmContextName.getNextNode(nextMibNode.name)

                except NoSuchInstanceError:
                    break

                self._contextMap[nextMibNode.syntax] = True

            self._contextBranchId = vacmContextName.branchVersionId

        # 3.2.1
        if contextName not in self._contextMap:
            raise error.StatusInformation(errorIndication=errind.noSuchContext)

        # Rebuild groupName map if changed

        (vacmGroupName,) = mibInstrumController.mibBuilder.importSymbols(
            "SNMP-VIEW-BASED-ACM-MIB", "vacmGroupName"
        )

        if self._groupNameBranchId != vacmGroupName.branchVersionId:
            (vacmSecurityToGroupEntry,) = mibInstrumController.mibBuilder.importSymbols(
                "SNMP-VIEW-BASED-ACM-MIB", "vacmSecurityToGroupEntry"
            )

            self._groupNameMap.clear()

            nextMibNode = vacmGroupName

            while True:
                try:
                    nextMibNode = vacmGroupName.getNextNode(nextMibNode.name)

                except NoSuchInstanceError:
                    break

                instId = nextMibNode.name[len(vacmGroupName.name) :]

                indices = vacmSecurityToGroupEntry.getIndicesFromInstId(instId)

                self._groupNameMap[indices] = nextMibNode.syntax

            self._groupNameBranchId = vacmGroupName.branchVersionId

        # 3.2.2
        indices = securityModel, securityName

        try:
            groupName = self._groupNameMap[indices]

        except KeyError:
            raise error.StatusInformation(errorIndication=errind.noGroupName)

        # Rebuild access map if changed

        (vacmAccessStatus,) = mibInstrumController.mibBuilder.importSymbols(
            "SNMP-VIEW-BASED-ACM-MIB", "vacmAccessStatus"
        )

        if self._accessBranchId != vacmAccessStatus.branchVersionId:
            (
                vacmAccessEntry,
                vacmAccessContextPrefix,
                vacmAccessSecurityModel,
                vacmAccessSecurityLevel,
                vacmAccessContextMatch,
                vacmAccessReadViewName,
                vacmAccessWriteViewName,
                vacmAccessNotifyViewName,
            ) = mibInstrumController.mibBuilder.importSymbols(
                "SNMP-VIEW-BASED-ACM-MIB",
                "vacmAccessEntry",
                "vacmAccessContextPrefix",
                "vacmAccessSecurityModel",
                "vacmAccessSecurityLevel",
                "vacmAccessContextMatch",
                "vacmAccessReadViewName",
                "vacmAccessWriteViewName",
                "vacmAccessNotifyViewName",
            )

            self._accessMap.clear()

            nextMibNode = vacmAccessStatus

            while True:
                try:
                    nextMibNode = vacmAccessStatus.getNextNode(nextMibNode.name)

                except NoSuchInstanceError:
                    break

                if nextMibNode.syntax != 1:  # active row
                    continue

                instId = nextMibNode.name[len(vacmAccessStatus.name) :]

                indices = vacmAccessEntry.getIndicesFromInstId(instId)

                vacmGroupName = indices[0]

                self._addAccessEntry(
                    vacmGroupName,
                    vacmAccessContextPrefix.getNode(
                        vacmAccessContextPrefix.name + instId
                    ).syntax,
                    vacmAccessSecurityModel.getNode(
                        vacmAccessSecurityModel.name + instId
                    ).syntax,
                    vacmAccessSecurityLevel.getNode(
                        vacmAccessSecurityLevel.name + instId
                    ).syntax,
                    vacmAccessContextMatch.getNode(
                        vacmAccessContextMatch.name + instId
                    ).syntax,
                    vacmAccessReadViewName.getNode(
                        vacmAccessReadViewName.name + instId
                    ).syntax,
                    vacmAccessWriteViewName.getNode(
                        vacmAccessWriteViewName.name + instId
                    ).syntax,
                    vacmAccessNotifyViewName.getNode(
                        vacmAccessNotifyViewName.name + instId
                    ).syntax,
                )

            self._accessBranchId = vacmAccessStatus.branchVersionId

        viewName = self._getFamilyViewName(
            groupName, contextName, securityModel, securityLevel, viewType
        )

        # Rebuild family subtree map if changed

        (vacmViewTreeFamilyViewName,) = mibInstrumController.mibBuilder.importSymbols(
            "SNMP-VIEW-BASED-ACM-MIB", "vacmViewTreeFamilyViewName"
        )

        if self._viewTreeBranchId != vacmViewTreeFamilyViewName.branchVersionId:
            (
                vacmViewTreeFamilySubtree,
                vacmViewTreeFamilyMask,
                vacmViewTreeFamilyType,
            ) = mibInstrumController.mibBuilder.importSymbols(
                "SNMP-VIEW-BASED-ACM-MIB",
                "vacmViewTreeFamilySubtree",
                "vacmViewTreeFamilyMask",
                "vacmViewTreeFamilyType",
            )

            self._viewTreeMap.clear()

            powerOfTwo = [2**exp for exp in range(7, -1, -1)]

            nextMibNode = vacmViewTreeFamilyViewName

            while True:
                try:
                    nextMibNode = vacmViewTreeFamilyViewName.getNextNode(
                        nextMibNode.name
                    )

                except NoSuchInstanceError:
                    break

                if nextMibNode.syntax not in self._viewTreeMap:
                    self._viewTreeMap[nextMibNode.syntax] = []

                instId = nextMibNode.name[len(vacmViewTreeFamilyViewName.name) :]

                subtree = vacmViewTreeFamilySubtree.getNode(
                    vacmViewTreeFamilySubtree.name + instId
                ).syntax

                mask = vacmViewTreeFamilyMask.getNode(
                    vacmViewTreeFamilyMask.name + instId
                ).syntax

                mode = vacmViewTreeFamilyType.getNode(
                    vacmViewTreeFamilyType.name + instId
                ).syntax

                mask = mask.asNumbers()
                maskLength = min(len(mask) * 8, len(subtree))

                ignoredSubOids = [
                    i * 8 + j
                    for i, octet in enumerate(mask)
                    for j, bit in enumerate(powerOfTwo)
                    if not (bit & octet) and i * 8 + j < maskLength
                ]

                if ignoredSubOids:
                    pattern = list(subtree)

                    for ignoredSubOid in ignoredSubOids:
                        pattern[ignoredSubOid] = 0

                    subtree = subtree.clone(pattern)

                entry = subtree, ignoredSubOids, mode == 1

                self._viewTreeMap[nextMibNode.syntax].append(entry)

            for entries in self._viewTreeMap.values():
                entries.sort(key=lambda x: (len(x[0]), x[0]))

            self._viewTreeBranchId = vacmViewTreeFamilyViewName.branchVersionId

        # 3.2.5a
        indices = viewName

        try:
            entries = self._viewTreeMap[indices]

        except KeyError:
            return error.StatusInformation(errorIndication=errind.notInView)

        accessAllowed = False

        for entry in entries:
            subtree, ignoredSubOids, included = entry

            if ignoredSubOids:
                subOids = list(variableName)

                for ignoredSubOid in ignoredSubOids:
                    subOids[ignoredSubOid] = 0

                normalizedVariableName = subtree.clone(subOids)

            else:
                normalizedVariableName = variableName

            if subtree.isPrefixOf(normalizedVariableName):
                accessAllowed = included

        # 3.2.5c
        if not accessAllowed:
            raise error.StatusInformation(errorIndication=errind.notInView)



# http://www.python.org/dev/peps/pep-0396/
__version__ = "6.0.13"
# another variable is required to prevent semantic release from updating version in more than one place
main_version = __version__
# backward compatibility
# for beta versions, integer casting throws an exception, so string part must be cut off
if "beta" in __version__:
    main_version = __version__.split("-beta")[0]
version = tuple(int(x) for x in main_version.split("."))
majorVersionId = version[0]



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2019, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
from pysnmp.proto import error
from pysnmp import nextid


class Cache:
    __stateReference = nextid.Integer(0xFFFFFF)
    __msgID = nextid.Integer(0xFFFFFF)

    def __init__(self):
        self.__msgIdIndex = {}
        self.__stateReferenceIndex = {}
        self.__sendPduHandleIdx = {}
        # Message expiration mechanics
        self.__expirationQueue = {}
        self.__expirationTimer = 0

    # Server mode cache handling

    def newStateReference(self):
        return self.__stateReference()

    def pushByStateRef(self, stateReference, **msgInfo):
        if stateReference in self.__stateReferenceIndex:
            raise error.ProtocolError(
                f"Cache dup for stateReference={stateReference} at {self}"
            )
        expireAt = self.__expirationTimer + 600
        self.__stateReferenceIndex[stateReference] = msgInfo, expireAt

        # Schedule to expire
        if expireAt not in self.__expirationQueue:
            self.__expirationQueue[expireAt] = {}
        if "stateReference" not in self.__expirationQueue[expireAt]:
            self.__expirationQueue[expireAt]["stateReference"] = {}
        self.__expirationQueue[expireAt]["stateReference"][stateReference] = 1

    def popByStateRef(self, stateReference):
        if stateReference in self.__stateReferenceIndex:
            cacheInfo = self.__stateReferenceIndex[stateReference]
        else:
            raise error.ProtocolError(
                f"Cache miss for stateReference={stateReference} at {self}"
            )
        del self.__stateReferenceIndex[stateReference]
        cacheEntry, expireAt = cacheInfo
        del self.__expirationQueue[expireAt]["stateReference"][stateReference]
        return cacheEntry

    # Client mode cache handling

    def newMsgID(self):
        return self.__msgID()

    def pushByMsgId(self, msgId, **msgInfo):
        if msgId in self.__msgIdIndex:
            raise error.ProtocolError(f"Cache dup for msgId={msgId} at {self}")
        expireAt = self.__expirationTimer + 600
        self.__msgIdIndex[msgId] = msgInfo, expireAt

        self.__sendPduHandleIdx[msgInfo["sendPduHandle"]] = msgId

        # Schedule to expire
        if expireAt not in self.__expirationQueue:
            self.__expirationQueue[expireAt] = {}
        if "msgId" not in self.__expirationQueue[expireAt]:
            self.__expirationQueue[expireAt]["msgId"] = {}
        self.__expirationQueue[expireAt]["msgId"][msgId] = 1

    def popByMsgId(self, msgId):
        if msgId in self.__msgIdIndex:
            cacheInfo = self.__msgIdIndex[msgId]
        else:
            raise error.ProtocolError(f"Cache miss for msgId={msgId} at {self}")
        msgInfo, expireAt = cacheInfo
        del self.__sendPduHandleIdx[msgInfo["sendPduHandle"]]
        del self.__msgIdIndex[msgId]
        cacheEntry, expireAt = cacheInfo
        del self.__expirationQueue[expireAt]["msgId"][msgId]
        return cacheEntry

    def popBySendPduHandle(self, sendPduHandle):
        if sendPduHandle in self.__sendPduHandleIdx:
            self.popByMsgId(self.__sendPduHandleIdx[sendPduHandle])

    def expireCaches(self):
        # Uses internal clock to expire pending messages
        if self.__expirationTimer in self.__expirationQueue:
            cacheInfo = self.__expirationQueue[self.__expirationTimer]
            if "stateReference" in cacheInfo:
                for stateReference in cacheInfo["stateReference"]:
                    del self.__stateReferenceIndex[stateReference]
            if "msgId" in cacheInfo:
                for msgId in cacheInfo["msgId"]:
                    del self.__msgIdIndex[msgId]
            del self.__expirationQueue[self.__expirationTimer]
        self.__expirationTimer += 1



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2019, Ilya Etingof <etingof@gmail.com>
#
# Copyright (c) 2024, LeXtudio Inc. <support@lextudio.com>
#
# License: https://www.pysnmp.com/pysnmp/license.html
#
import time
import sys
from pysnmp.proto.secmod.base import AbstractSecurityModel
from pysnmp.proto.secmod.rfc3414.auth import hmacmd5, hmacsha, noauth
from pysnmp.proto.secmod.rfc3414.priv import des, nopriv
from pysnmp.proto.secmod.rfc3826.priv import aes
from pysnmp.proto.secmod.rfc7860.auth import hmacsha2
from pysnmp.proto.secmod.eso.priv import des3, aes192, aes256
from pysnmp.smi.error import NoSuchInstanceError
from pysnmp.proto import api, rfc1155, rfc3411, errind, error
from pysnmp import debug
from pyasn1.type import univ, namedtype, constraint
from pyasn1.codec.ber import encoder, decoder, eoo
from pyasn1.error import PyAsn1Error
from pyasn1.compat.octets import null

# API to rfc1905 protocol objects
pMod = api.protoModules[api.protoVersion2c]


# USM security params


class UsmSecurityParameters(rfc1155.TypeCoercionHackMixIn, univ.Sequence):
    componentType = namedtype.NamedTypes(
        namedtype.NamedType("msgAuthoritativeEngineId", univ.OctetString()),
        namedtype.NamedType(
            "msgAuthoritativeEngineBoots",
            univ.Integer().subtype(
                subtypeSpec=constraint.ValueRangeConstraint(0, 2147483647)
            ),
        ),
        namedtype.NamedType(
            "msgAuthoritativeEngineTime",
            univ.Integer().subtype(
                subtypeSpec=constraint.ValueRangeConstraint(0, 2147483647)
            ),
        ),
        namedtype.NamedType(
            "msgUserName",
            univ.OctetString().subtype(
                subtypeSpec=constraint.ValueSizeConstraint(0, 32)
            ),
        ),
        namedtype.NamedType("msgAuthenticationParameters", univ.OctetString()),
        namedtype.NamedType("msgPrivacyParameters", univ.OctetString()),
    )


class SnmpUSMSecurityModel(AbstractSecurityModel):
    securityModelID = 3
    authServices = {
        hmacmd5.HmacMd5.serviceID: hmacmd5.HmacMd5(),
        hmacsha.HmacSha.serviceID: hmacsha.HmacSha(),
        hmacsha2.HmacSha2.sha224ServiceID: hmacsha2.HmacSha2(
            hmacsha2.HmacSha2.sha224ServiceID
        ),
        hmacsha2.HmacSha2.sha256ServiceID: hmacsha2.HmacSha2(
            hmacsha2.HmacSha2.sha256ServiceID
        ),
        hmacsha2.HmacSha2.sha384ServiceID: hmacsha2.HmacSha2(
            hmacsha2.HmacSha2.sha384ServiceID
        ),
        hmacsha2.HmacSha2.sha512ServiceID: hmacsha2.HmacSha2(
            hmacsha2.HmacSha2.sha512ServiceID
        ),
        noauth.NoAuth.serviceID: noauth.NoAuth(),
    }
    privServices = {
        des.Des.serviceID: des.Des(),
        des3.Des3.serviceID: des3.Des3(),
        aes.Aes.serviceID: aes.Aes(),
        aes192.AesBlumenthal192.serviceID: aes192.AesBlumenthal192(),
        aes256.AesBlumenthal256.serviceID: aes256.AesBlumenthal256(),
        aes192.Aes192.serviceID: aes192.Aes192(),  # non-standard
        aes256.Aes256.serviceID: aes256.Aes256(),  # non-standard
        nopriv.NoPriv.serviceID: nopriv.NoPriv(),
    }

    # If this, normally impossible, SNMP engine ID is present in LCD, we will use
    # its master/localized keys when preparing SNMP message towards any unknown peer
    # SNMP engine
    wildcardSecurityEngineId = pMod.OctetString(hexValue="0000000000")

    def __init__(self):
        AbstractSecurityModel.__init__(self)
        self.__securityParametersSpec = UsmSecurityParameters()
        self.__timeline = {}
        self.__timelineExpQueue = {}
        self.__expirationTimer = 0
        self.__paramsBranchId = -1

    def __sec2usr(self, snmpEngine, securityName, securityEngineID=None):
        mibBuilder = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder
        (usmUserEngineID,) = mibBuilder.importSymbols(
            "SNMP-USER-BASED-SM-MIB", "usmUserEngineID"
        )
        if self.__paramsBranchId != usmUserEngineID.branchVersionId:
            usmUserName, usmUserSecurityName = mibBuilder.importSymbols(
                "SNMP-USER-BASED-SM-MIB", "usmUserName", "usmUserSecurityName"
            )

            self.__securityToUserMap = {}

            nextMibNode = usmUserEngineID

            while True:
                try:
                    nextMibNode = usmUserEngineID.getNextNode(nextMibNode.name)

                except NoSuchInstanceError:
                    self.__paramsBranchId = usmUserEngineID.branchVersionId
                    debug.logger & debug.flagSM and debug.logger(
                        "_sec2usr: built snmpEngineId + securityName to userName map, version {}: {!r}".format(
                            self.__paramsBranchId, self.__securityToUserMap
                        )
                    )
                    break

                instId = nextMibNode.name[len(usmUserSecurityName.name) :]

                __engineID = usmUserEngineID.getNode(
                    usmUserEngineID.name + instId
                ).syntax
                __userName = usmUserName.getNode(usmUserName.name + instId).syntax
                __securityName = usmUserSecurityName.getNode(
                    usmUserSecurityName.name + instId
                ).syntax

                k = __engineID, __securityName

                # first (lesser) securityName wins
                if k not in self.__securityToUserMap:
                    self.__securityToUserMap[k] = __userName

        if securityEngineID is None:
            (snmpEngineID,) = mibBuilder.importSymbols(
                "__SNMP-FRAMEWORK-MIB", "snmpEngineID"
            )
            securityEngineID = snmpEngineID.syntax

        try:
            userName = self.__securityToUserMap[(securityEngineID, securityName)]
        except KeyError:
            debug.logger & debug.flagSM and debug.logger(
                f"_sec2usr: no entry exists for snmpEngineId {securityEngineID!r}, securityName {securityName!r}"
            )
            raise NoSuchInstanceError()  # emulate MIB lookup

        debug.logger & debug.flagSM and debug.logger(
            "_sec2usr: using userName {!r} for snmpEngineId {!r}, securityName {!r}".format(
                userName, securityEngineID, securityName
            )
        )

        return userName

    @staticmethod
    def __getUserInfo(mibInstrumController, securityEngineID, userName):
        (usmUserEntry,) = mibInstrumController.mibBuilder.importSymbols(
            "SNMP-USER-BASED-SM-MIB", "usmUserEntry"
        )
        tblIdx = usmUserEntry.getInstIdFromIndices(securityEngineID, userName)
        # Get userName & securityName
        usmUserName = usmUserEntry.getNode(usmUserEntry.name + (2,) + tblIdx).syntax
        usmUserSecurityName = usmUserEntry.getNode(
            usmUserEntry.name + (3,) + tblIdx
        ).syntax
        # Get protocols
        usmUserAuthProtocol = usmUserEntry.getNode(
            usmUserEntry.name + (5,) + tblIdx
        ).syntax
        usmUserPrivProtocol = usmUserEntry.getNode(
            usmUserEntry.name + (8,) + tblIdx
        ).syntax
        # Get keys
        (pysnmpUsmKeyEntry,) = mibInstrumController.mibBuilder.importSymbols(
            "PYSNMP-USM-MIB", "pysnmpUsmKeyEntry"
        )
        pysnmpUsmKeyAuthLocalized = pysnmpUsmKeyEntry.getNode(
            pysnmpUsmKeyEntry.name + (1,) + tblIdx
        ).syntax
        pysnmpUsmKeyPrivLocalized = pysnmpUsmKeyEntry.getNode(
            pysnmpUsmKeyEntry.name + (2,) + tblIdx
        ).syntax
        return (
            usmUserName,
            usmUserSecurityName,
            usmUserAuthProtocol,
            pysnmpUsmKeyAuthLocalized,
            usmUserPrivProtocol,
            pysnmpUsmKeyPrivLocalized,
        )

    def __cloneUserInfo(self, mibInstrumController, securityEngineID, userName):
        (snmpEngineID,) = mibInstrumController.mibBuilder.importSymbols(
            "__SNMP-FRAMEWORK-MIB", "snmpEngineID"
        )
        # Proto entry
        (usmUserEntry,) = mibInstrumController.mibBuilder.importSymbols(
            "SNMP-USER-BASED-SM-MIB", "usmUserEntry"
        )
        tblIdx1 = usmUserEntry.getInstIdFromIndices(snmpEngineID.syntax, userName)
        # Get proto protocols
        usmUserName = usmUserEntry.getNode(usmUserEntry.name + (2,) + tblIdx1)
        usmUserSecurityName = usmUserEntry.getNode(usmUserEntry.name + (3,) + tblIdx1)
        usmUserCloneFrom = usmUserEntry.getNode(usmUserEntry.name + (4,) + tblIdx1)
        usmUserAuthProtocol = usmUserEntry.getNode(usmUserEntry.name + (5,) + tblIdx1)
        usmUserPrivProtocol = usmUserEntry.getNode(usmUserEntry.name + (8,) + tblIdx1)
        # Get proto keys
        (pysnmpUsmKeyEntry,) = mibInstrumController.mibBuilder.importSymbols(
            "PYSNMP-USM-MIB", "pysnmpUsmKeyEntry"
        )
        pysnmpUsmKeyAuth = pysnmpUsmKeyEntry.getNode(
            pysnmpUsmKeyEntry.name + (3,) + tblIdx1
        )
        pysnmpUsmKeyPriv = pysnmpUsmKeyEntry.getNode(
            pysnmpUsmKeyEntry.name + (4,) + tblIdx1
        )

        # Create new row from proto values

        tblIdx2 = usmUserEntry.getInstIdFromIndices(securityEngineID, userName)

        # New row
        mibInstrumController.writeVars(((usmUserEntry.name + (13,) + tblIdx2, 4),))

        # Set user&securityNames
        usmUserEntry.getNode(
            usmUserEntry.name + (2,) + tblIdx2
        ).syntax = usmUserName.syntax
        usmUserEntry.getNode(
            usmUserEntry.name + (3,) + tblIdx2
        ).syntax = usmUserSecurityName.syntax

        # Store a reference to original row
        usmUserEntry.getNode(
            usmUserEntry.name + (4,) + tblIdx2
        ).syntax = usmUserCloneFrom.syntax.clone(tblIdx1)

        # Set protocols
        usmUserEntry.getNode(
            usmUserEntry.name + (5,) + tblIdx2
        ).syntax = usmUserAuthProtocol.syntax
        usmUserEntry.getNode(
            usmUserEntry.name + (8,) + tblIdx2
        ).syntax = usmUserPrivProtocol.syntax

        # Localize and set keys
        (pysnmpUsmKeyEntry,) = mibInstrumController.mibBuilder.importSymbols(
            "PYSNMP-USM-MIB", "pysnmpUsmKeyEntry"
        )
        pysnmpUsmKeyAuthLocalized = pysnmpUsmKeyEntry.getNode(
            pysnmpUsmKeyEntry.name + (1,) + tblIdx2
        )
        if usmUserAuthProtocol.syntax in self.authServices:
            localizeKey = self.authServices[usmUserAuthProtocol.syntax].localizeKey
            localAuthKey = localizeKey(pysnmpUsmKeyAuth.syntax, securityEngineID)
        else:
            raise error.StatusInformation(
                errorIndication=errind.unsupportedAuthProtocol
            )
        if localAuthKey is not None:
            pysnmpUsmKeyAuthLocalized.syntax = pysnmpUsmKeyAuthLocalized.syntax.clone(
                localAuthKey
            )
        pysnmpUsmKeyPrivLocalized = pysnmpUsmKeyEntry.getNode(
            pysnmpUsmKeyEntry.name + (2,) + tblIdx2
        )
        if usmUserPrivProtocol.syntax in self.privServices:
            localizeKey = self.privServices[usmUserPrivProtocol.syntax].localizeKey
            localPrivKey = localizeKey(
                usmUserAuthProtocol.syntax, pysnmpUsmKeyPriv.syntax, securityEngineID
            )
        else:
            raise error.StatusInformation(
                errorIndication=errind.unsupportedPrivProtocol
            )
        if localPrivKey is not None:
            pysnmpUsmKeyPrivLocalized.syntax = pysnmpUsmKeyPrivLocalized.syntax.clone(
                localPrivKey
            )
        return (
            usmUserName.syntax,
            usmUserSecurityName.syntax,
            usmUserAuthProtocol.syntax,
            pysnmpUsmKeyAuthLocalized.syntax,
            usmUserPrivProtocol.syntax,
            pysnmpUsmKeyPrivLocalized.syntax,
        )

    def __generateRequestOrResponseMsg(
        self,
        snmpEngine,
        messageProcessingModel,
        globalData,
        maxMessageSize,
        securityModel,
        securityEngineID,
        securityName,
        securityLevel,
        scopedPDU,
        securityStateReference,
        ctx,
    ):
        mibBuilder = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder
        snmpEngineID = mibBuilder.importSymbols("__SNMP-FRAMEWORK-MIB", "snmpEngineID")[
            0
        ].syntax
        msg = globalData

        # 3.1.1
        if securityStateReference is not None:
            # 3.1.1a
            cachedSecurityData = self._cache.pop(securityStateReference)
            usmUserName = cachedSecurityData["msgUserName"]
            if "usmUserSecurityName" in cachedSecurityData:
                usmUserSecurityName = cachedSecurityData["usmUserSecurityName"]
            else:
                usmUserSecurityName = usmUserName
            if "usmUserAuthProtocol" in cachedSecurityData:
                usmUserAuthProtocol = cachedSecurityData["usmUserAuthProtocol"]
            else:
                usmUserAuthProtocol = noauth.NoAuth.serviceID
            if "usmUserAuthKeyLocalized" in cachedSecurityData:
                usmUserAuthKeyLocalized = cachedSecurityData["usmUserAuthKeyLocalized"]
            else:
                usmUserAuthKeyLocalized = None
            if "usmUserPrivProtocol" in cachedSecurityData:
                usmUserPrivProtocol = cachedSecurityData["usmUserPrivProtocol"]
            else:
                usmUserPrivProtocol = nopriv.NoPriv.serviceID
            if "usmUserPrivKeyLocalized" in cachedSecurityData:
                usmUserPrivKeyLocalized = cachedSecurityData["usmUserPrivKeyLocalized"]
            else:
                usmUserPrivKeyLocalized = None

            securityEngineID = snmpEngineID

            debug.logger & debug.flagSM and debug.logger(
                "__generateRequestOrResponseMsg: using cached USM user entry "
                'usmUserName "%s" '
                'usmUserSecurityName "%s" '
                'usmUserAuthProtocol "%s" '
                'usmUserAuthKeyLocalized "%s" '
                'usmUserPrivProtocol "%s" '
                'usmUserPrivKeyLocalized "%s" for '
                'securityEngineID "%s" and  securityName "%s" found by '
                'securityStateReference "%s" '
                % (
                    usmUserName,
                    usmUserSecurityName,
                    usmUserAuthProtocol,
                    usmUserAuthKeyLocalized and usmUserAuthKeyLocalized.prettyPrint(),
                    usmUserPrivProtocol,
                    usmUserPrivKeyLocalized and usmUserPrivKeyLocalized.prettyPrint(),
                    securityEngineID.prettyPrint(),
                    securityName,
                    securityStateReference,
                )
            )

        elif securityEngineID:
            # 3.1.1b
            try:
                try:
                    (
                        usmUserName,
                        usmUserSecurityName,
                        usmUserAuthProtocol,
                        usmUserAuthKeyLocalized,
                        usmUserPrivProtocol,
                        usmUserPrivKeyLocalized,
                    ) = self.__getUserInfo(
                        snmpEngine.msgAndPduDsp.mibInstrumController,
                        securityEngineID,
                        self.__sec2usr(snmpEngine, securityName, securityEngineID),
                    )

                except NoSuchInstanceError:
                    (
                        usmUserName,
                        usmUserSecurityName,
                        usmUserAuthProtocol,
                        usmUserAuthKeyLocalized,
                        usmUserPrivProtocol,
                        usmUserPrivKeyLocalized,
                    ) = self.__getUserInfo(
                        snmpEngine.msgAndPduDsp.mibInstrumController,
                        self.wildcardSecurityEngineId,
                        self.__sec2usr(
                            snmpEngine, securityName, self.wildcardSecurityEngineId
                        ),
                    )

                debug.logger & debug.flagSM and debug.logger(
                    "__generateRequestOrResponseMsg: found USM user entry "
                    'usmUserName "%s" '
                    'usmUserSecurityName "%s" '
                    'usmUserAuthProtocol "%s" '
                    'usmUserAuthKeyLocalized "%s" '
                    'usmUserPrivProtocol "%s" '
                    'usmUserPrivKeyLocalized "%s" by '
                    'securityEngineID "%s" and  securityName "%s"'
                    % (
                        usmUserName,
                        usmUserSecurityName,
                        usmUserAuthProtocol,
                        usmUserAuthKeyLocalized
                        and usmUserAuthKeyLocalized.prettyPrint(),
                        usmUserPrivProtocol,
                        usmUserPrivKeyLocalized
                        and usmUserPrivKeyLocalized.prettyPrint(),
                        securityEngineID.prettyPrint(),
                        securityName,
                    )
                )

            except NoSuchInstanceError:
                (pysnmpUsmDiscovery,) = mibBuilder.importSymbols(
                    "__PYSNMP-USM-MIB", "pysnmpUsmDiscovery"
                )
                reportUnknownName = not pysnmpUsmDiscovery.syntax
                if not reportUnknownName:
                    try:
                        (
                            usmUserName,
                            usmUserSecurityName,
                            usmUserAuthProtocol,
                            usmUserAuthKeyLocalized,
                            usmUserPrivProtocol,
                            usmUserPrivKeyLocalized,
                        ) = self.__cloneUserInfo(
                            snmpEngine.msgAndPduDsp.mibInstrumController,
                            securityEngineID,
                            self.__sec2usr(snmpEngine, securityName),
                        )

                        debug.logger & debug.flagSM and debug.logger(
                            "__generateRequestOrResponseMsg: cloned USM user entry "
                            'usmUserName "%s" '
                            'usmUserSecurityName "%s" '
                            'usmUserAuthProtocol "%s" '
                            'usmUserAuthKeyLocalized "%s" '
                            'usmUserPrivProtocol "%s" '
                            'usmUserPrivKeyLocalized "%s" for '
                            'securityEngineID "%s" and  securityName "%s"'
                            % (
                                usmUserName,
                                usmUserSecurityName,
                                usmUserAuthProtocol,
                                usmUserAuthKeyLocalized
                                and usmUserAuthKeyLocalized.prettyPrint(),
                                usmUserPrivProtocol,
                                usmUserPrivKeyLocalized
                                and usmUserPrivKeyLocalized.prettyPrint(),
                                securityEngineID.prettyPrint(),
                                securityName,
                            )
                        )

                    except NoSuchInstanceError:
                        debug.logger & debug.flagSM and debug.logger(
                            "__generateRequestOrResponseMsg: failed to clone "
                            'USM user for securityEngineID "%s" securityName '
                            '"%s"' % (securityEngineID, securityName)
                        )

                        reportUnknownName = True

                if reportUnknownName:
                    raise error.StatusInformation(
                        errorIndication=errind.unknownSecurityName
                    )

            except PyAsn1Error:
                debug.logger & debug.flagSM and debug.logger(
                    f"__generateRequestOrResponseMsg: {sys.exc_info()[1]}"
                )
                (snmpInGenErrs,) = mibBuilder.importSymbols(
                    "__SNMPv2-MIB", "snmpInGenErrs"
                )
                snmpInGenErrs.syntax += 1
                raise error.StatusInformation(errorIndication=errind.invalidMsg)

        else:
            # 4. (start SNMP engine ID discovery)
            securityEngineID = securityName = null
            securityLevel = 1

            scopedPDU.setComponentByPosition(
                0,
                null,
                verifyConstraints=False,
                matchTags=False,
                matchConstraints=False,
            )

            headerData = msg.getComponentByPosition(1)

            # Clear possible auth&priv flags
            headerData.setComponentByPosition(
                2,
                univ.OctetString(hexValue="04"),
                verifyConstraints=False,
                matchTags=False,
                matchConstraints=False,
            )

            emptyPdu = scopedPDU.getComponentByPosition(2).getComponent()

            # we edit the rest of the structures in-place because they
            # are ours for as long as this stack lasts, however PDU
            # is more persistent and should not be touched

            emptyPdu = emptyPdu.clone()
            pMod.apiPDU.setDefaults(emptyPdu)

            scopedPDU.getComponentByPosition(2).setComponentByType(
                emptyPdu.tagSet,
                emptyPdu,
                verifyConstraints=False,
                matchTags=False,
                matchConstraints=False,
            )

            usmUserName = usmUserSecurityName = null
            usmUserAuthProtocol = noauth.NoAuth.serviceID
            usmUserPrivProtocol = nopriv.NoPriv.serviceID
            usmUserAuthKeyLocalized = usmUserPrivKeyLocalized = None

            debug.logger & debug.flagSM and debug.logger(
                "__generateRequestOrResponseMsg: using blank USM info for peer "
                "SNMP engine ID discovery "
                'usmUserName "%s" '
                'usmUserSecurityName "%s" '
                'usmUserAuthProtocol "%s" '
                'usmUserAuthKeyLocalized "%s" '
                'usmUserPrivProtocol "%s" '
                'usmUserPrivKeyLocalized "%s" for '
                'securityEngineID "%s" and  securityName "%s"'
                % (
                    usmUserName,
                    usmUserSecurityName,
                    usmUserAuthProtocol,
                    usmUserAuthKeyLocalized,
                    usmUserPrivProtocol,
                    usmUserPrivKeyLocalized,
                    securityEngineID and securityEngineID.prettyPrint(),
                    securityName,
                )
            )

        # 3.1.2
        if ctx:
            if (
                ctx == errind.authenticationFailure
                or ctx == errind.unknownSecurityName
                or ctx == errind.decryptionError
            ):
                debug.logger & debug.flagSM and debug.logger(
                    "__generateRequestOrResponseMsg: ctx is known error indication %s"
                    % ctx
                )
                securityLevel = 1
                headerData = msg.getComponentByPosition(1)

                # Clear possible auth&priv flags
                headerData.setComponentByPosition(
                    2,
                    univ.OctetString(hexValue="04"),
                    verifyConstraints=False,
                    matchTags=False,
                    matchConstraints=False,
                )

        if securityLevel == 3:
            if (
                usmUserAuthProtocol == noauth.NoAuth.serviceID
                or usmUserPrivProtocol == nopriv.NoPriv.serviceID
            ):
                raise error.StatusInformation(
                    errorIndication=errind.unsupportedSecurityLevel
                )

        # 3.1.3
        if securityLevel == 3 or securityLevel == 2:
            if usmUserAuthProtocol == noauth.NoAuth.serviceID:
                raise error.StatusInformation(
                    errorIndication=errind.unsupportedSecurityLevel
                )

        securityParameters = self.__securityParametersSpec

        scopedPDUData = msg.setComponentByPosition(3).getComponentByPosition(3)
        scopedPDUData.setComponentByPosition(
            0,
            scopedPDU,
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
        )

        # 3.1.6a
        snmpEngineBoots = snmpEngineTime = 0

        if securityLevel in (1, 2, 3):
            pdu = scopedPDU.getComponentByPosition(2).getComponent()

            # 3.1.6.b
            if pdu.tagSet in rfc3411.unconfirmedClassPDUs:
                (snmpEngineBoots, snmpEngineTime) = mibBuilder.importSymbols(
                    "__SNMP-FRAMEWORK-MIB", "snmpEngineBoots", "snmpEngineTime"
                )

                snmpEngineBoots = snmpEngineBoots.syntax
                snmpEngineTime = snmpEngineTime.syntax.clone()

                debug.logger & debug.flagSM and debug.logger(
                    "__generateRequestOrResponseMsg: read snmpEngineBoots, snmpEngineTime from LCD"
                )

            # 3.1.6a
            elif securityEngineID in self.__timeline:
                (
                    snmpEngineBoots,
                    snmpEngineTime,
                    latestReceivedEngineTime,
                    latestUpdateTimestamp,
                ) = self.__timeline[securityEngineID]

                debug.logger & debug.flagSM and debug.logger(
                    "__generateRequestOrResponseMsg: read snmpEngineBoots, snmpEngineTime from timeline"
                )

            # 3.1.6.c
            else:
                debug.logger & debug.flagSM and debug.logger(
                    "__generateRequestOrResponseMsg: assuming zero snmpEngineBoots, snmpEngineTime"
                )

            debug.logger & debug.flagSM and debug.logger(
                "__generateRequestOrResponseMsg: use snmpEngineBoots {} snmpEngineTime {} for securityEngineID {!r}".format(
                    snmpEngineBoots, snmpEngineTime, securityEngineID
                )
            )

        # 3.1.4a
        if securityLevel == 3:
            if usmUserPrivProtocol in self.privServices:
                privHandler = self.privServices[usmUserPrivProtocol]
            else:
                raise error.StatusInformation(errorIndication=errind.encryptionError)

            debug.logger & debug.flagSM and debug.logger(
                "__generateRequestOrResponseMsg: scopedPDU %s" % scopedPDU.prettyPrint()
            )

            try:
                dataToEncrypt = encoder.encode(scopedPDU)

            except PyAsn1Error:
                debug.logger & debug.flagSM and debug.logger(
                    "__generateRequestOrResponseMsg: scopedPDU serialization error: %s"
                    % sys.exc_info()[1]
                )
                raise error.StatusInformation(errorIndication=errind.serializationError)

            debug.logger & debug.flagSM and debug.logger(
                "__generateRequestOrResponseMsg: scopedPDU encoded into %s"
                % debug.hexdump(dataToEncrypt)
            )

            # noinspection PyUnboundLocalVariable
            (encryptedData, privParameters) = privHandler.encryptData(
                usmUserPrivKeyLocalized,
                (snmpEngineBoots, snmpEngineTime, None),
                dataToEncrypt,
            )

            securityParameters.setComponentByPosition(
                5,
                privParameters,
                verifyConstraints=False,
                matchTags=False,
                matchConstraints=False,
            )
            scopedPDUData.setComponentByPosition(
                1,
                encryptedData,
                verifyConstraints=False,
                matchTags=False,
                matchConstraints=False,
            )

            debug.logger & debug.flagSM and debug.logger(
                "__generateRequestOrResponseMsg: scopedPDU ciphered into %s"
                % debug.hexdump(encryptedData)
            )

        # 3.1.4b
        elif securityLevel == 1 or securityLevel == 2:
            securityParameters.setComponentByPosition(5, "")

        debug.logger & debug.flagSM and debug.logger(
            "__generateRequestOrResponseMsg: %s" % scopedPDUData.prettyPrint()
        )

        # 3.1.5
        securityParameters.setComponentByPosition(
            0,
            securityEngineID,
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
        )
        securityParameters.setComponentByPosition(
            1,
            snmpEngineBoots,
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
        )
        securityParameters.setComponentByPosition(
            2,
            snmpEngineTime,
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
        )

        # 3.1.7
        securityParameters.setComponentByPosition(
            3,
            usmUserName,
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
        )

        # 3.1.8a
        if securityLevel == 3 or securityLevel == 2:
            if usmUserAuthProtocol in self.authServices:
                authHandler = self.authServices[usmUserAuthProtocol]
            else:
                raise error.StatusInformation(
                    errorIndication=errind.authenticationFailure
                )

            # extra-wild hack to facilitate BER substrate in-place re-write
            securityParameters.setComponentByPosition(
                4, "\x00" * authHandler.digestLength
            )

            debug.logger & debug.flagSM and debug.logger(
                f"__generateRequestOrResponseMsg: {securityParameters.prettyPrint()}"
            )

            try:
                msg.setComponentByPosition(
                    2, encoder.encode(securityParameters), verifyConstraints=False
                )

            except PyAsn1Error:
                debug.logger & debug.flagSM and debug.logger(
                    "__generateRequestOrResponseMsg: securityParameters serialization error: %s"
                    % sys.exc_info()[1]
                )
                raise error.StatusInformation(errorIndication=errind.serializationError)

            debug.logger & debug.flagSM and debug.logger(
                "__generateRequestOrResponseMsg: auth outgoing msg: %s"
                % msg.prettyPrint()
            )

            try:
                wholeMsg = encoder.encode(msg)

            except PyAsn1Error:
                debug.logger & debug.flagSM and debug.logger(
                    "__generateRequestOrResponseMsg: msg serialization error: %s"
                    % sys.exc_info()[1]
                )
                raise error.StatusInformation(errorIndication=errind.serializationError)

            # noinspection PyUnboundLocalVariable
            authenticatedWholeMsg = authHandler.authenticateOutgoingMsg(
                usmUserAuthKeyLocalized, wholeMsg
            )

        # 3.1.8b
        else:
            securityParameters.setComponentByPosition(
                4, "", verifyConstraints=False, matchTags=False, matchConstraints=False
            )

            debug.logger & debug.flagSM and debug.logger(
                f"__generateRequestOrResponseMsg: {securityParameters.prettyPrint()}"
            )

            try:
                msg.setComponentByPosition(
                    2,
                    encoder.encode(securityParameters),
                    verifyConstraints=False,
                    matchTags=False,
                    matchConstraints=False,
                )

            except PyAsn1Error:
                debug.logger & debug.flagSM and debug.logger(
                    "__generateRequestOrResponseMsg: secutiryParameters serialization error: %s"
                    % sys.exc_info()[1]
                )
                raise error.StatusInformation(errorIndication=errind.serializationError)

            try:
                debug.logger & debug.flagSM and debug.logger(
                    "__generateRequestOrResponseMsg: plain outgoing msg: %s"
                    % msg.prettyPrint()
                )
                authenticatedWholeMsg = encoder.encode(msg)

            except PyAsn1Error:
                debug.logger & debug.flagSM and debug.logger(
                    "__generateRequestOrResponseMsg: msg serialization error: %s"
                    % sys.exc_info()[1]
                )
                raise error.StatusInformation(errorIndication=errind.serializationError)

        debug.logger & debug.flagSM and debug.logger(
            "__generateRequestOrResponseMsg: {} outgoing msg: {}".format(
                securityLevel > 1 and "authenticated" or "plain",
                debug.hexdump(authenticatedWholeMsg),
            )
        )

        # 3.1.9
        return msg.getComponentByPosition(2), authenticatedWholeMsg

    def generateRequestMsg(
        self,
        snmpEngine,
        messageProcessingModel,
        globalData,
        maxMessageSize,
        securityModel,
        securityEngineID,
        securityName,
        securityLevel,
        scopedPDU,
    ):
        return self.__generateRequestOrResponseMsg(
            snmpEngine,
            messageProcessingModel,
            globalData,
            maxMessageSize,
            securityModel,
            securityEngineID,
            securityName,
            securityLevel,
            scopedPDU,
            None,
            None,
        )

    def generateResponseMsg(
        self,
        snmpEngine,
        messageProcessingModel,
        globalData,
        maxMessageSize,
        securityModel,
        securityEngineID,
        securityName,
        securityLevel,
        scopedPDU,
        securityStateReference,
        ctx,
    ):
        return self.__generateRequestOrResponseMsg(
            snmpEngine,
            messageProcessingModel,
            globalData,
            maxMessageSize,
            securityModel,
            securityEngineID,
            securityName,
            securityLevel,
            scopedPDU,
            securityStateReference,
            ctx,
        )

    # 3.2
    def processIncomingMsg(
        self,
        snmpEngine,
        messageProcessingModel,
        maxMessageSize,
        securityParameters,
        securityModel,
        securityLevel,
        wholeMsg,
        msg,
    ):
        mibBuilder = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder

        # 3.2.9 -- moved up here to be able to report
        # maxSizeResponseScopedPDU on error
        # (48 - maximum SNMPv3 header length)
        maxSizeResponseScopedPDU = int(maxMessageSize) - len(securityParameters) - 48

        debug.logger & debug.flagSM and debug.logger(
            "processIncomingMsg: securityParameters %s"
            % debug.hexdump(securityParameters)
        )

        # 3.2.1
        securityParameters, rest = decoder.decode(
            securityParameters, asn1Spec=self.__securityParametersSpec
        )

        debug.logger & debug.flagSM and debug.logger(
            f"processIncomingMsg: {securityParameters.prettyPrint()}"
        )

        if eoo.endOfOctets.isSameTypeWith(securityParameters):
            raise error.StatusInformation(errorIndication=errind.parseError)

        # 3.2.2
        msgAuthoritativeEngineId = securityParameters.getComponentByPosition(0)
        securityStateReference = self._cache.push(
            msgUserName=securityParameters.getComponentByPosition(3)
        )

        debug.logger & debug.flagSM and debug.logger(
            "processIncomingMsg: cache write securityStateReference {} by msgUserName {}".format(
                securityStateReference, securityParameters.getComponentByPosition(3)
            )
        )

        scopedPduData = msg.getComponentByPosition(3)

        # Used for error reporting
        contextEngineId = mibBuilder.importSymbols(
            "__SNMP-FRAMEWORK-MIB", "snmpEngineID"
        )[0].syntax
        contextName = null

        snmpEngineID = mibBuilder.importSymbols("__SNMP-FRAMEWORK-MIB", "snmpEngineID")[
            0
        ].syntax

        # 3.2.3
        if (
            msgAuthoritativeEngineId != snmpEngineID
            and msgAuthoritativeEngineId not in self.__timeline
        ):
            if msgAuthoritativeEngineId and 4 < len(msgAuthoritativeEngineId) < 33:
                # 3.2.3a - cloned user when request was sent
                debug.logger & debug.flagSM and debug.logger(
                    f"processIncomingMsg: non-synchronized securityEngineID {msgAuthoritativeEngineId!r}"
                )
            else:
                # 3.2.3b
                debug.logger & debug.flagSM and debug.logger(
                    "processIncomingMsg: peer requested snmpEngineID discovery"
                )
                (usmStatsUnknownEngineIDs,) = mibBuilder.importSymbols(
                    "__SNMP-USER-BASED-SM-MIB", "usmStatsUnknownEngineIDs"
                )
                usmStatsUnknownEngineIDs.syntax += 1
                debug.logger & debug.flagSM and debug.logger(
                    "processIncomingMsg: null or malformed msgAuthoritativeEngineId"
                )
                (pysnmpUsmDiscoverable,) = mibBuilder.importSymbols(
                    "__PYSNMP-USM-MIB", "pysnmpUsmDiscoverable"
                )
                if pysnmpUsmDiscoverable.syntax:
                    debug.logger & debug.flagSM and debug.logger(
                        "processIncomingMsg: starting snmpEngineID discovery procedure"
                    )

                    # Report original contextName
                    if scopedPduData.getName() != "plaintext":
                        debug.logger & debug.flagSM and debug.logger(
                            "processIncomingMsg: scopedPduData not plaintext %s"
                            % scopedPduData.prettyPrint()
                        )
                        raise error.StatusInformation(
                            errorIndication=errind.unknownEngineID
                        )

                    # 7.2.6.a.1
                    scopedPdu = scopedPduData.getComponent()
                    contextEngineId = scopedPdu.getComponentByPosition(0)
                    contextName = scopedPdu.getComponentByPosition(1)

                    raise error.StatusInformation(
                        errorIndication=errind.unknownEngineID,
                        oid=usmStatsUnknownEngineIDs.name,
                        val=usmStatsUnknownEngineIDs.syntax,
                        securityStateReference=securityStateReference,
                        securityLevel=securityLevel,
                        contextEngineId=contextEngineId,
                        contextName=contextName,
                        scopedPDU=scopedPdu,
                        maxSizeResponseScopedPDU=maxSizeResponseScopedPDU,
                    )
                else:
                    debug.logger & debug.flagSM and debug.logger(
                        "processIncomingMsg: will not discover EngineID"
                    )
                    # free securityStateReference XXX
                    raise error.StatusInformation(
                        errorIndication=errind.unknownEngineID
                    )

        msgUserName = securityParameters.getComponentByPosition(3)

        debug.logger & debug.flagSM and debug.logger(
            "processIncomingMsg: read from securityParams msgAuthoritativeEngineId {!r} msgUserName {!r}".format(
                msgAuthoritativeEngineId, msgUserName
            )
        )

        if msgUserName:
            # 3.2.4
            try:
                (
                    usmUserName,
                    usmUserSecurityName,
                    usmUserAuthProtocol,
                    usmUserAuthKeyLocalized,
                    usmUserPrivProtocol,
                    usmUserPrivKeyLocalized,
                ) = self.__getUserInfo(
                    snmpEngine.msgAndPduDsp.mibInstrumController,
                    msgAuthoritativeEngineId,
                    msgUserName,
                )
                debug.logger & debug.flagSM and debug.logger(
                    "processIncomingMsg: read user info from LCD"
                )

            except NoSuchInstanceError:
                try:
                    (
                        usmUserName,
                        usmUserSecurityName,
                        usmUserAuthProtocol,
                        usmUserAuthKeyLocalized,
                        usmUserPrivProtocol,
                        usmUserPrivKeyLocalized,
                    ) = self.__getUserInfo(
                        snmpEngine.msgAndPduDsp.mibInstrumController,
                        self.wildcardSecurityEngineId,
                        msgUserName,
                    )
                    debug.logger & debug.flagSM and debug.logger(
                        "processIncomingMsg: read wildcard user info from LCD"
                    )

                except NoSuchInstanceError:
                    debug.logger & debug.flagSM and debug.logger(
                        "processIncomingMsg: unknown securityEngineID {!r} msgUserName {!r}".format(
                            msgAuthoritativeEngineId, msgUserName
                        )
                    )

                    (usmStatsUnknownUserNames,) = mibBuilder.importSymbols(
                        "__SNMP-USER-BASED-SM-MIB", "usmStatsUnknownUserNames"
                    )
                    usmStatsUnknownUserNames.syntax += 1

                    raise error.StatusInformation(
                        errorIndication=errind.unknownSecurityName,
                        oid=usmStatsUnknownUserNames.name,
                        val=usmStatsUnknownUserNames.syntax,
                        securityStateReference=securityStateReference,
                        securityLevel=securityLevel,
                        contextEngineId=contextEngineId,
                        contextName=contextName,
                        msgUserName=msgUserName,
                        maxSizeResponseScopedPDU=maxSizeResponseScopedPDU,
                    )

            except PyAsn1Error:
                debug.logger & debug.flagSM and debug.logger(
                    f"processIncomingMsg: {sys.exc_info()[1]}"
                )
                (snmpInGenErrs,) = mibBuilder.importSymbols(
                    "__SNMPv2-MIB", "snmpInGenErrs"
                )
                snmpInGenErrs.syntax += 1
                raise error.StatusInformation(errorIndication=errind.invalidMsg)
        else:
            # empty username used for engineID discovery
            usmUserName = usmUserSecurityName = null
            usmUserAuthProtocol = noauth.NoAuth.serviceID
            usmUserPrivProtocol = nopriv.NoPriv.serviceID
            usmUserAuthKeyLocalized = usmUserPrivKeyLocalized = None

        debug.logger & debug.flagSM and debug.logger(
            "processIncomingMsg: now have usmUserName {!r} usmUserSecurityName {!r} usmUserAuthProtocol {!r} usmUserPrivProtocol {!r} for msgUserName {!r}".format(
                usmUserName,
                usmUserSecurityName,
                usmUserAuthProtocol,
                usmUserPrivProtocol,
                msgUserName,
            )
        )

        # 3.2.11 (moved up here to let Reports be authenticated & encrypted)
        self._cache.pop(securityStateReference)
        securityStateReference = self._cache.push(
            msgUserName=securityParameters.getComponentByPosition(3),
            usmUserSecurityName=usmUserSecurityName,
            usmUserAuthProtocol=usmUserAuthProtocol,
            usmUserAuthKeyLocalized=usmUserAuthKeyLocalized,
            usmUserPrivProtocol=usmUserPrivProtocol,
            usmUserPrivKeyLocalized=usmUserPrivKeyLocalized,
        )

        msgAuthoritativeEngineBoots = securityParameters.getComponentByPosition(1)
        msgAuthoritativeEngineTime = securityParameters.getComponentByPosition(2)

        snmpEngine.observer.storeExecutionContext(
            snmpEngine,
            "rfc3414.processIncomingMsg",
            dict(
                securityEngineId=msgAuthoritativeEngineId,
                snmpEngineBoots=msgAuthoritativeEngineBoots,
                snmpEngineTime=msgAuthoritativeEngineTime,
                userName=usmUserName,
                securityName=usmUserSecurityName,
                authProtocol=usmUserAuthProtocol,
                authKey=usmUserAuthKeyLocalized,
                privProtocol=usmUserPrivProtocol,
                privKey=usmUserPrivKeyLocalized,
            ),
        )
        snmpEngine.observer.clearExecutionContext(
            snmpEngine, "rfc3414.processIncomingMsg"
        )

        # 3.2.5
        if msgAuthoritativeEngineId == snmpEngineID:
            # Authoritative SNMP engine: make sure securityLevel is sufficient
            badSecIndication = None
            if securityLevel == 3:
                if usmUserAuthProtocol == noauth.NoAuth.serviceID:
                    badSecIndication = "authPriv wanted while auth not expected"
                if usmUserPrivProtocol == nopriv.NoPriv.serviceID:
                    badSecIndication = "authPriv wanted while priv not expected"
            elif securityLevel == 2:
                if usmUserAuthProtocol == noauth.NoAuth.serviceID:
                    badSecIndication = "authNoPriv wanted while auth not expected"
                if usmUserPrivProtocol != nopriv.NoPriv.serviceID:
                    # 4 (discovery phase always uses authenticated messages)
                    if msgAuthoritativeEngineBoots or msgAuthoritativeEngineTime:
                        badSecIndication = "authNoPriv wanted while priv expected"

            elif securityLevel == 1:
                if usmUserAuthProtocol != noauth.NoAuth.serviceID:
                    badSecIndication = "noAuthNoPriv wanted while auth expected"
                if usmUserPrivProtocol != nopriv.NoPriv.serviceID:
                    badSecIndication = "noAuthNoPriv wanted while priv expected"
            if badSecIndication:
                (usmStatsUnsupportedSecLevels,) = mibBuilder.importSymbols(
                    "__SNMP-USER-BASED-SM-MIB", "usmStatsUnsupportedSecLevels"
                )
                usmStatsUnsupportedSecLevels.syntax += 1
                debug.logger & debug.flagSM and debug.logger(
                    "processIncomingMsg: reporting inappropriate security level for user {}: {}".format(
                        msgUserName, badSecIndication
                    )
                )
                raise error.StatusInformation(
                    errorIndication=errind.unsupportedSecurityLevel,
                    oid=usmStatsUnsupportedSecLevels.name,
                    val=usmStatsUnsupportedSecLevels.syntax,
                    securityStateReference=securityStateReference,
                    securityLevel=securityLevel,
                    contextEngineId=contextEngineId,
                    contextName=contextName,
                    msgUserName=msgUserName,
                    maxSizeResponseScopedPDU=maxSizeResponseScopedPDU,
                )

        # 3.2.6
        if securityLevel in (1, 2, 3):
            if usmUserName:
                if usmUserAuthProtocol in self.authServices:
                    authHandler = self.authServices[usmUserAuthProtocol]
                else:
                    raise error.StatusInformation(
                        errorIndication=errind.authenticationFailure
                    )

                hash = securityParameters.getComponentByPosition(4)
                try:
                    authHandler.authenticateIncomingMsg(
                        usmUserAuthKeyLocalized, hash, wholeMsg
                    )

                except error.StatusInformation:
                    if (
                        len(hash) != 0
                    ):  # don't throw error if hash is empty (and agent returned REPORT)
                        (usmStatsWrongDigests,) = mibBuilder.importSymbols(
                            "__SNMP-USER-BASED-SM-MIB", "usmStatsWrongDigests"
                        )
                        usmStatsWrongDigests.syntax += 1
                        raise error.StatusInformation(
                            errorIndication=errind.authenticationFailure,
                            oid=usmStatsWrongDigests.name,
                            val=usmStatsWrongDigests.syntax,
                            securityStateReference=securityStateReference,
                            securityLevel=securityLevel,
                            contextEngineId=contextEngineId,
                            contextName=contextName,
                            msgUserName=msgUserName,
                            maxSizeResponseScopedPDU=maxSizeResponseScopedPDU,
                        )

                debug.logger & debug.flagSM and debug.logger(
                    "processIncomingMsg: incoming msg authenticated"
                )

            # synchronize time with authed peer
            self.__timeline[msgAuthoritativeEngineId] = (
                securityParameters.getComponentByPosition(1),
                securityParameters.getComponentByPosition(2),
                securityParameters.getComponentByPosition(2),
                int(time.time()),
            )

            timerResolution = (
                snmpEngine.transportDispatcher is None
                and 1.0
                or snmpEngine.transportDispatcher.getTimerResolution()
            )
            expireAt = int(self.__expirationTimer + 300 / timerResolution)
            if expireAt not in self.__timelineExpQueue:
                self.__timelineExpQueue[expireAt] = []
            self.__timelineExpQueue[expireAt].append(msgAuthoritativeEngineId)

            debug.logger & debug.flagSM and debug.logger(
                f"processIncomingMsg: store timeline for securityEngineID {msgAuthoritativeEngineId!r}"
            )

        # 3.2.7
        if securityLevel == 3 or securityLevel == 2:
            if msgAuthoritativeEngineId == snmpEngineID:
                # Authoritative SNMP engine: use local notion (SF bug #1649032)
                (snmpEngineBoots, snmpEngineTime) = mibBuilder.importSymbols(
                    "__SNMP-FRAMEWORK-MIB", "snmpEngineBoots", "snmpEngineTime"
                )
                snmpEngineBoots = snmpEngineBoots.syntax
                snmpEngineTime = snmpEngineTime.syntax.clone()
                idleTime = 0
                debug.logger & debug.flagSM and debug.logger(
                    "processIncomingMsg: read snmpEngineBoots ({}), snmpEngineTime ({}) from LCD".format(
                        snmpEngineBoots, snmpEngineTime
                    )
                )
            else:
                # Non-authoritative SNMP engine: use cached estimates
                if msgAuthoritativeEngineId in self.__timeline:
                    (
                        snmpEngineBoots,
                        snmpEngineTime,
                        latestReceivedEngineTime,
                        latestUpdateTimestamp,
                    ) = self.__timeline[msgAuthoritativeEngineId]
                    # time passed since last talk with this SNMP engine
                    idleTime = int(time.time()) - latestUpdateTimestamp
                    debug.logger & debug.flagSM and debug.logger(
                        "processIncomingMsg: read timeline snmpEngineBoots {} snmpEngineTime {} for msgAuthoritativeEngineId {!r}, idle time {} secs".format(
                            snmpEngineBoots,
                            snmpEngineTime,
                            msgAuthoritativeEngineId,
                            idleTime,
                        )
                    )
                else:
                    raise error.ProtocolError("Peer SNMP engine info missing")

            # 3.2.7a
            if msgAuthoritativeEngineId == snmpEngineID:
                if (
                    snmpEngineBoots == 2147483647
                    or snmpEngineBoots != msgAuthoritativeEngineBoots
                    or abs(
                        idleTime + int(snmpEngineTime) - int(msgAuthoritativeEngineTime)
                    )
                    > 150
                ):
                    (usmStatsNotInTimeWindows,) = mibBuilder.importSymbols(
                        "__SNMP-USER-BASED-SM-MIB", "usmStatsNotInTimeWindows"
                    )
                    usmStatsNotInTimeWindows.syntax += 1
                    raise error.StatusInformation(
                        errorIndication=errind.notInTimeWindow,
                        oid=usmStatsNotInTimeWindows.name,
                        val=usmStatsNotInTimeWindows.syntax,
                        securityStateReference=securityStateReference,
                        securityLevel=2,
                        contextEngineId=contextEngineId,
                        contextName=contextName,
                        msgUserName=msgUserName,
                        maxSizeResponseScopedPDU=maxSizeResponseScopedPDU,
                    )
            # 3.2.7b
            else:
                # 3.2.7b.1
                # noinspection PyUnboundLocalVariable
                if (
                    msgAuthoritativeEngineBoots > snmpEngineBoots
                    or msgAuthoritativeEngineBoots == snmpEngineBoots
                    and msgAuthoritativeEngineTime > latestReceivedEngineTime
                ):
                    self.__timeline[msgAuthoritativeEngineId] = (
                        msgAuthoritativeEngineBoots,
                        msgAuthoritativeEngineTime,
                        msgAuthoritativeEngineTime,
                        int(time.time()),
                    )

                    timerResolution = (
                        snmpEngine.transportDispatcher is None
                        and 1.0
                        or snmpEngine.transportDispatcher.getTimerResolution()
                    )
                    expireAt = int(self.__expirationTimer + 300 / timerResolution)
                    if expireAt not in self.__timelineExpQueue:
                        self.__timelineExpQueue[expireAt] = []
                    self.__timelineExpQueue[expireAt].append(msgAuthoritativeEngineId)

                    debug.logger & debug.flagSM and debug.logger(
                        "processIncomingMsg: stored timeline msgAuthoritativeEngineBoots {} msgAuthoritativeEngineTime {} for msgAuthoritativeEngineId {!r}".format(
                            msgAuthoritativeEngineBoots,
                            msgAuthoritativeEngineTime,
                            msgAuthoritativeEngineId,
                        )
                    )

                # 3.2.7b.2
                if (
                    snmpEngineBoots == 2147483647
                    or msgAuthoritativeEngineBoots < snmpEngineBoots
                    or msgAuthoritativeEngineBoots == snmpEngineBoots
                    and abs(
                        idleTime + int(snmpEngineTime) - int(msgAuthoritativeEngineTime)
                    )
                    > 150
                ):
                    raise error.StatusInformation(
                        errorIndication=errind.notInTimeWindow, msgUserName=msgUserName
                    )

        # 3.2.8a
        if securityLevel == 3:
            if usmUserPrivProtocol in self.privServices:
                privHandler = self.privServices[usmUserPrivProtocol]
            else:
                raise error.StatusInformation(
                    errorIndication=errind.decryptionError, msgUserName=msgUserName
                )
            encryptedPDU = scopedPduData.getComponentByPosition(1)
            if encryptedPDU is None:  # no ciphertext
                raise error.StatusInformation(
                    errorIndication=errind.decryptionError, msgUserName=msgUserName
                )

            try:
                decryptedData = privHandler.decryptData(
                    usmUserPrivKeyLocalized,
                    (
                        securityParameters.getComponentByPosition(1),
                        securityParameters.getComponentByPosition(2),
                        securityParameters.getComponentByPosition(5),
                    ),
                    encryptedPDU,
                )
                debug.logger & debug.flagSM and debug.logger(
                    "processIncomingMsg: PDU deciphered into %s"
                    % debug.hexdump(decryptedData)
                )

            except error.StatusInformation:
                (usmStatsDecryptionErrors,) = mibBuilder.importSymbols(
                    "__SNMP-USER-BASED-SM-MIB", "usmStatsDecryptionErrors"
                )
                usmStatsDecryptionErrors.syntax += 1
                raise error.StatusInformation(
                    errorIndication=errind.decryptionError,
                    oid=usmStatsDecryptionErrors.name,
                    val=usmStatsDecryptionErrors.syntax,
                    securityStateReference=securityStateReference,
                    securityLevel=securityLevel,
                    contextEngineId=contextEngineId,
                    contextName=contextName,
                    msgUserName=msgUserName,
                    maxSizeResponseScopedPDU=maxSizeResponseScopedPDU,
                )
            scopedPduSpec = scopedPduData.setComponentByPosition(
                0
            ).getComponentByPosition(0)
            try:
                scopedPDU, rest = decoder.decode(decryptedData, asn1Spec=scopedPduSpec)

            except PyAsn1Error:
                debug.logger & debug.flagSM and debug.logger(
                    "processIncomingMsg: scopedPDU decoder failed %s"
                    % sys.exc_info()[0]
                )
                (usmStatsDecryptionErrors,) = mibBuilder.importSymbols(
                    "__SNMP-USER-BASED-SM-MIB", "usmStatsDecryptionErrors"
                )
                usmStatsDecryptionErrors.syntax += 1
                raise error.StatusInformation(
                    errorIndication=errind.decryptionError,
                    oid=usmStatsDecryptionErrors.name,
                    val=usmStatsDecryptionErrors.syntax,
                    securityStateReference=securityStateReference,
                    securityLevel=securityLevel,
                    contextEngineId=contextEngineId,
                    contextName=contextName,
                    msgUserName=msgUserName,
                    maxSizeResponseScopedPDU=maxSizeResponseScopedPDU,
                )

            if eoo.endOfOctets.isSameTypeWith(scopedPDU):
                raise error.StatusInformation(
                    errorIndication=errind.decryptionError, msgUserName=msgUserName
                )
        else:
            # 3.2.8b
            scopedPDU = scopedPduData.getComponentByPosition(0)
            if scopedPDU is None:  # no plaintext
                raise error.StatusInformation(
                    errorIndication=errind.decryptionError, msgUserName=msgUserName
                )

        debug.logger & debug.flagSM and debug.logger(
            "processIncomingMsg: scopedPDU decoded %s" % scopedPDU.prettyPrint()
        )

        # 3.2.10
        securityName = usmUserSecurityName

        debug.logger & debug.flagSM and debug.logger(
            "processIncomingMsg: cached msgUserName {} info by securityStateReference {}".format(
                msgUserName, securityStateReference
            )
        )

        # Delayed to include details
        if not msgUserName and not msgAuthoritativeEngineId:
            (usmStatsUnknownUserNames,) = mibBuilder.importSymbols(
                "__SNMP-USER-BASED-SM-MIB", "usmStatsUnknownUserNames"
            )
            usmStatsUnknownUserNames.syntax += 1
            raise error.StatusInformation(
                errorIndication=errind.unknownSecurityName,
                oid=usmStatsUnknownUserNames.name,
                val=usmStatsUnknownUserNames.syntax,
                securityStateReference=securityStateReference,
                securityEngineID=msgAuthoritativeEngineId,
                securityLevel=securityLevel,
                contextEngineId=contextEngineId,
                contextName=contextName,
                msgUserName=msgUserName,
                maxSizeResponseScopedPDU=maxSizeResponseScopedPDU,
                PDU=scopedPDU,
            )

        # 3.2.12
        return (
            msgAuthoritativeEngineId,
            securityName,
            scopedPDU,
            maxSizeResponseScopedPDU,
            securityStateReference,
        )

    def __expireTimelineInfo(self):
        if self.__expirationTimer in self.__timelineExpQueue:
            for engineIdKey in self.__timelineExpQueue[self.__expirationTimer]:
                if engineIdKey in self.__timeline:
                    del self.__timeline[engineIdKey]
                    debug.logger & debug.flagSM and debug.logger(
                        f"__expireTimelineInfo: expiring {engineIdKey!r}"
                    )
            del self.__timelineExpQueue[self.__expirationTimer]
        self.__expirationTimer += 1

    def receiveTimerTick(self, snmpEngine, timeNow):
        self.__expireTimelineInfo()
